from .aniDBmapper import AniDBMapper


class ResponseResolver:
    def __init__(self, data):
        self.restag, self.rescode, self.resstr, self.datalines = self.parse(data)

    @staticmethod
    def parse(data):
        resline = data.split("\n", 1)[0]
        lines = data.split("\n")[1:-1]

        rescode, resstr = resline.split(" ", 1)
        if rescode[0] == "T":
            restag = rescode
            rescode, resstr = resstr.split(" ", 1)
        else:
            restag = None

        datalines = []
        for line in lines:
            datalines.append(line.split("|"))

        return restag, rescode, resstr, datalines

    def resolve(self, cmd):
        return responses[self.rescode](cmd, self.restag, self.rescode, self.resstr, self.datalines)


class Response:
    def __init__(self, cmd, restag, rescode, resstr, rawlines):
        self.req = cmd
        self.restag = restag
        self.rescode = rescode
        self.resstr = resstr
        self.rawlines = rawlines
        self.maper = AniDBMapper()

        self.attrs = None
        self.datalines = None

        self.codetail = ()
        self.codehead = ()
        self.coderep = ()

    def __repr__(self):
        tmp = "%s(%s,%s,%s) %s\n" % (self.__class__.__name__, repr(self.restag), repr(self.rescode), repr(self.resstr), repr(self.attrs))

        m = 0
        for line in self.datalines:
            for k, v in line.items():
                if len(k) > m:
                    m = len(k)

        for line in self.datalines:
            tmp += "  Line:\n"
            for k, v in line.items():
                tmp += "    %s:%s %s\n" % (k, (m - len(k)) * " ", v)
        return tmp

    def parse(self):
        tmp = self.resstr.split(" ", len(self.codehead))
        self.attrs = dict(zip(self.codehead, tmp[:-1]))
        self.resstr = tmp[-1]

        self.datalines = []
        for rawline in self.rawlines:
            normal = dict(zip(self.codetail, rawline))
            rawline = rawline[len(self.codetail) :]
            rep = []
            if self.coderep:
                while rawline:
                    tmp = dict(zip(self.coderep, rawline))
                    rawline = rawline[len(self.coderep) :]
                    rep.append(tmp)
            # normal['rep']=rep
            self.datalines.append(normal)

    def handle(self):
        if self.req:
            self.req.handle(self)


class LoginAcceptedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:
        sesskey - session key
        address - your address (ip:port) as seen by the server

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "LOGIN_ACCEPTED"
        self.codetail = ()
        self.coderep = ()

        nat = cmd.parameters["nat"]
        nat = int(nat or "0")
        if nat:
            self.codehead = ("sesskey", "address")
        else:
            self.codehead = ("sesskey",)


class LoginAcceptedNewVerResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:
        sesskey - session key
        address - your address (ip:port) as seen by the server

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "LOGIN_ACCEPTED_NEW_VER"
        self.codetail = ()
        self.coderep = ()

        nat = cmd.parameters["nat"]
        nat = int(nat or "0")
        if nat:
            self.codehead = ("sesskey", "address")
        else:
            self.codehead = ("sesskey",)


class LoggedOutResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "LOGGED_OUT"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class ResourceResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "RESOURCE"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class StatsResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "STATS"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class TopResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "TOP"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class UptimeResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        uptime - udpserver uptime in milliseconds

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "UPTIME"
        self.codehead = ()
        self.codetail = ("uptime",)
        self.coderep = ()


class EncryptionEnabledResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:
        salt - salt

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "ENCRYPTION_ENABLED"
        self.codehead = ("salt",)
        self.codetail = ()
        self.coderep = ()


class MylistEntryAddedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        entrycnt - number of entries added

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "MYLIST_ENTRY_ADDED"
        self.codehead = ()
        self.codetail = ("entrycnt",)
        self.coderep = ()


class MylistEntryDeletedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        entrycnt - number of entries

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "MYLIST_ENTRY_DELETED"
        self.codehead = ()
        self.codetail = ("entrycnt",)
        self.coderep = ()


class AddedFileResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "ADDED_FILE"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class AddedStreamResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "ADDED_STREAM"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class EncodingChangedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "ENCODING_CHANGED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class FileResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        eid - episode id
        gid - group id
        lid - mylist id
        state - state
        size - size
        ed2k - ed2k
        md5 - md5
        sha1 - sha1
        crc32 - crc32
        dublang - dub language
        sublang - sub language
        quality - quality
        source - source
        audiocodec - audio codec
        audiobitrate - audio bitrate
        videocodec - video codec
        videobitrate - video bitrate
        resolution - video resolution
        filetype - file type (extension)
        length - length in seconds
        description - description
        filename - anidb file name
        gname - group name
        gshortname - group short name
        epno - number of episode
        epname - ep english name
        epromaji - ep romaji name
        epkanji - ep kanji name
        totaleps - anime total episodes
        lastep - last episode nr (highest, not special)
        year - year
        type - type
        romaji - romaji name
        kanji - kanji name
        name - english name
        othername - other name
        shortnames - short name list
        synonyms - synonym list
        categories - category list
        relatedaids - related aid list
        producernames - producer name list
        producerids - producer id list

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "FILE"
        self.codehead = ()
        self.coderep = ()

        fmask = cmd.parameters["fmask"]
        amask = cmd.parameters["amask"]

        codeListF = self.maper.getFileCodesF(fmask)
        codeListA = self.maper.getFileCodesA(amask)
        # print "File - codelistF: "+str(codeListF)
        # print "File - codelistA: "+str(codeListA)

        self.codetail = tuple(["fid"] + codeListF + codeListA)


class MylistResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        lid - mylist id
        fid - file id
        eid - episode id
        aid - anime id
        gid - group id
        date - date when you added this to mylist
        state - the location of the file
        viewdate - date when you marked this watched
        storage - for example the title of the cd you have this on
        source - where you got the file (bittorrent,dc++,ed2k,...)
        other - other data regarding this file
        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "MYLIST"
        self.codehead = ()
        self.codetail = ("lid", "fid", "eid", "aid", "gid", "date", "state", "viewdate", "storage", "source", "other")
        self.coderep = ()


class MylistStatsResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        animes - animes
        eps - eps
        files - files
        filesizes - size of files
        animesadded - added animes
        epsadded - added eps
        filesadded - added files
        groupsadded - added groups
        leechperc - leech %
        lameperc - lame %
        viewedofdb - viewed % of db
        mylistofdb - mylist % of db
        viewedofmylist - viewed % of mylist
        viewedeps - number of viewed eps
        votes - votes
        reviews - reviews

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "MYLIST_STATS"
        self.codehead = ()
        self.codetail = (
            "animes",
            "eps",
            "files",
            "filesizes",
            "animesadded",
            "epsadded",
            "filesadded",
            "groupsadded",
            "leechperc",
            "lameperc",
            "viewedofdb",
            "mylistofdb",
            "viewedofmylist",
            "viewedeps",
            "votes",
            "reviews",
        )
        self.coderep = ()


class NotificationResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        nid         - notification id
        unknown     - unsure what this parameter is
        type        - Type of notification: type=>  0=all, 1=new, 2=group, 3=complete
        aid         - anime id
        priority    - low = 0, medium = 1, high = 2 (unconfirmed)
        date        - date notification subscribed to

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NOTIFICATION"
        self.codehead = ()
        self.codetail = ("nid", "unknown", "type", "aid", "priority", "date")
        self.coderep = ()


class AnimeResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "ANIME"
        self.codehead = ()
        self.coderep = ()

        # TODO: impl random anime
        amask = cmd.parameters["amask"]
        codeList = self.maper.getAnimeCodesA(amask)
        self.codetail = tuple(codeList)


class AnimeBestMatchResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "ANIME_BEST_MATCH"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class RandomanimeResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "RANDOMANIME"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class EpisodeResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        eid - episode id
        aid - anime id
        length - length
        rating - rating
        votes - votes
        epno - number of episode
        name - english name of episode
        romaji - romaji name of episode
        kanji - kanji name of episode

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "EPISODE"
        self.codehead = ()
        self.codetail = ("eid", "aid", "length", "rating", "votes", "epno", "name", "romaji", "kanji")
        self.coderep = ()


class ProducerResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        pid - producer id
        name - name of producer
        shortname - short name
        othername - other name
        type - type
        pic - picture name
        url - home page url

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "PRODUCER"
        self.codehead = ()
        self.codetail = ("pid", "name", "shortname", "othername", "type", "pic", "url")
        self.coderep = ()


class GroupResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        gid - group id
        rating - rating
        votes - votes
        animes - anime count
        files - file count
        name - name
        shortname  - short
        ircchannel - irc channel
        ircserver  - irc server
        url - url

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "GROUP"
        self.codehead = ()
        self.codetail = ("gid", "rating", "votes", "animes", "files", "name", "shortname", "ircchannel", "ircserver", "url")
        self.coderep = ()


class GroupstatusResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        gid - group id
        rating - rating
        votes - votes
        animes - anime count
        files - file count
        name - name
        shortname  - short
        ircchannel - irc channel
        ircserver  - irc server
        url - url

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "GROUPSTATUS"
        self.codehead = ()
        self.codetail = ("gid", "name", "state", " last_episode_number", "rating", "votes", "episode_range")
        self.coderep = ()


class BuddyListResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:
        start - mylist entry number of first buddy on this packet
        end - mylist entry number of last buddy on this packet
        total - total number of buddies on mylist

        data:
        uid - uid
        name - username
        state - state

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "BUDDY_LIST"
        self.codehead = ("start", "end", "total")
        self.codetail = ("uid", "username", "state")
        self.coderep = ()


class BuddyStateResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:
        start - mylist entry number of first buddy on this packet
        end - mylist entry number of last buddy on this packet
        total - total number of buddies on mylist

        data:
        uid - uid
        state - online state

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "BUDDY_STATE"
        self.codehead = ("start", "end", "total")
        self.codetail = ("uid", "state")
        self.coderep = ()


class BuddyAddedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "BUDDY_ADDED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class BuddyDeletedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "BUDDY_DELETED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class BuddyAcceptedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "BUDDY_ACCEPTED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class BuddyDeniedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "BUDDY_DENIED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class VotedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        name - aname/ename/gname

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "VOTED"
        self.codehead = ()
        self.codetail = ("name",)
        self.coderep = ()


class VoteFoundResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        name - aname/ename/gname
        value - vote value

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "VOTE_FOUND"
        self.codehead = ()
        self.codetail = ("name", "value")
        self.coderep = ()


class VoteUpdatedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        name - aname/ename/gname
        value - vote value

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "VOTE_UPDATED"
        self.codehead = ()
        self.codetail = ("name", "value")
        self.coderep = ()


class VoteRevokedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        name - aname/ename/gname
        value - vote value

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "VOTE_REVOKED"
        self.codehead = ()
        self.codetail = ("name", "value")
        self.coderep = ()


class NotificationAddedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        nid - notofication id

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NOTIFICATION_ITEM_ADDED"
        self.codehead = ()
        self.codetail = ("nid",)
        self.coderep = ()


class NotificationDeletedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        nid - notofication id

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NOTIFICATION_ITEM_DELETED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NotificationUpdatedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        nid - notofication id

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NOTIFICATION_ITEM_UPDATED"
        self.codehead = ()
        self.codetail = ("nid",)
        self.coderep = ()


class MultipleNotificationResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        nid         - notification id
        type        -
        priority    -


        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "MULTIPLE NOTIFICATION ITEMS"
        self.codehead = ()
        self.codetail = ()
        self.coderep = (
            "nid",
            "type",
            "priority",
        )


class NotificationEnabledResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NOTIFICATION_ENABLED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NotificationNotifyResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:
        nid - notify packet id

        data:
        aid - anime id
        date - date
        count - count
        name - name of the anime

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NOTIFICATION_NOTIFY"
        self.codehead = ("nid",)
        self.codetail = ("aid", "date", "count", "name")
        self.coderep = ()


class NotificationMessageResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:
        nid - notify packet id

        data:
        type - type
        date - date
        uid - user id of the sender
        name - name of the sender
        subject - subject

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NOTIFICATION_MESSAGE"
        self.codehead = ("nid",)
        self.codetail = ("type", "date", "uid", "name", "subject")
        self.coderep = ()


class NotificationBuddyResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:
        nid - notify packet id

        data:
        uid - buddy uid
        type - event type

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NOTIFICATION_BUDDY"
        self.codehead = ("notify_packet_id",)
        self.codetail = ("uid", "type")
        self.coderep = ()


class NotificationShutdownResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:
        nid - notify packet id

        data:
        time - time offline
        comment - comment

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NOTIFICATION_SHUTDOWN"
        self.codehead = ("nid",)
        self.codetail = ("time", "comment")
        self.coderep = ()


class PushackConfirmedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "PUSHACK_CONFIRMED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NotifyackSuccessfulMResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NOTIFYACK_SUCCESSFUL_M"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NotifyackSuccessfulNResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NOTIFYACK_SUCCESSFUL_N"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NotificationStateResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        notifies - pending notifies
        msgs - pending msgs
        buddys - number of online buddys

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NOTIFICATION"
        self.codehead = ()
        self.coderep = ()

        buddy = cmd.parameters["buddy"]
        buddy = int(buddy or "0")
        if buddy:
            self.codetail = ("notifies", "msgs", "buddys")
        else:
            self.codetail = ("notifies", "msgs")


class NotifylistResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        type - type
        nid - notify id

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NOTIFYLIST"
        self.codehead = ()
        self.codetail = ("type", "nid")
        self.coderep = ()


class NotifygetMessageResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        nid - notify id
        uid - from user id
        uname - from username
        date - date
        type - type
        title - title
        body - body

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NOTIFYGET_MESSAGE"
        self.codehead = ()
        self.codetail = ("nid", "uid", "uname", "date", "type", "title", "body")
        self.coderep = ()


class NotifygetNotifyResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        aid - aid
        type - type
        count - count
        date - date
        name - anime name
        fid     - comma separated list of fids
        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NOTIFYGET_NOTIFY"
        self.codehead = ()
        self.codetail = ("aid", "type", "count", "date", "name", "fid")
        self.coderep = ()


class SendmsgSuccessfulResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "SENDMSG_SUCCESSFUL"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class UserResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        uid - user id

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "USER"
        self.codehead = ()
        self.codetail = ("uid",)
        self.coderep = ()


class PongResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "PONG"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class AuthpongResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "AUTHPONG"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchResourceResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_SUCH_RESOURCE"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class ApiPasswordNotDefinedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "API_PASSWORD_NOT_DEFINED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class FileAlreadyInMylistResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "FILE_ALREADY_IN_MYLIST"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class MylistEntryEditedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        entries - number of entries edited

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "MYLIST_ENTRY_EDITED"
        self.codehead = ()
        self.codetail = ("entries",)
        self.coderep = ()


class MultipleMylistEntriesResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        name - anime title
        eps - episodes
        unknowneps - eps with state unknown
        hddeps - eps with state on hdd
        cdeps - eps with state on cd
        deletedeps - eps with state deleted
        watchedeps - watched eps
        gshortname - group short name
        geps - eps for group

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "MULTIPLE_MYLIST_ENTRIES"
        self.codehead = ()
        self.codetail = ("name", "eps", "unknowneps", "hddeps", "cdeps", "deletedeps", "watchedeps")
        self.coderep = ("gshortname", "geps")


class SizeHashExistsResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "SIZE_HASH_EXISTS"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class InvalidDataResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "INVALID_DATA"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class StreamnoidUsedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "STREAMNOID_USED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchFileResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_SUCH_FILE"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchEntryResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_SUCH_ENTRY"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class MultipleFilesFoundResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        fid - file id

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "MULTIPLE_FILES_FOUND"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ("fid",)


class NoSuchNotificationResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_SUCH_NOTIFICATION"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoGroupsFoundResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO GROUPS FOUND"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchAnimeResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_SUCH_ANIME"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchEpisodeResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_SUCH_EPISODE"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchProducerResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_SUCH_PRODUCER"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchGroupResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_SUCH_GROUP"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class BuddyAlreadyAddedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "BUDDY_ALREADY_ADDED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchBuddyResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_SUCH_BUDDY"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class BuddyAlreadyAcceptedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "BUDDY_ALREADY_ACCEPTED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class BuddyAlreadyDeniedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "BUDDY_ALREADY_DENIED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchVoteResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_SUCH_VOTE"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class InvalidVoteTypeResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "INVALID_VOTE_TYPE"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class InvalidVoteValueResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "INVALID_VOTE_VALUE"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class PermvoteNotAllowedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        aname - name of the anime

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "PERMVOTE_NOT_ALLOWED"
        self.codehead = ()
        self.codetail = ("aname",)
        self.coderep = ()


class AlreadyPermvotedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        name - aname/ename/gname

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "ALREADY_PERMVOTED"
        self.codehead = ()
        self.codetail = ("name",)
        self.coderep = ()


class NotificationDisabledResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NOTIFICATION_DISABLED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchPacketPendingResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_SUCH_PACKET_PENDING"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchEntryMResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_SUCH_ENTRY_M"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchEntryNResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_SUCH_ENTRY_N"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchMessageResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_SUCH_MESSAGE"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchNotifyResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_SUCH_NOTIFY"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchUserResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_SUCH_USER"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoChanges(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_CHANGES"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NotLoggedInResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NOT_LOGGED_IN"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchMylistFileResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_SUCH_MYLIST_FILE"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchMylistEntryResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_SUCH_MYLIST_ENTRY"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class LoginFailedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "LOGIN_FAILED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class LoginFirstResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "LOGIN_FIRST"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class AccessDeniedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "ACCESS_DENIED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class ClientVersionOutdatedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "CLIENT_VERSION_OUTDATED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class ClientBannedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "CLIENT_BANNED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class IllegalInputOrAccessDeniedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "ILLEGAL_INPUT_OR_ACCESS_DENIED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class InvalidSessionResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "INVALID_SESSION"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchEncryptionTypeResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "NO_SUCH_ENCRYPTION_TYPE"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class EncodingNotSupportedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "ENCODING_NOT_SUPPORTED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class BannedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "BANNED"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class UnknownCommandResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "UNKNOWN_COMMAND"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class InternalServerErrorResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "INTERNAL_SERVER_ERROR"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class AnidbOutOfServiceResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "ANIDB_OUT_OF_SERVICE"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class ServerBusyResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "SERVER_BUSY"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class ApiViolationResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "API_VIOLATION"
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class VersionResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        version - server version

        """
        super().__init__(cmd, restag, rescode, resstr, datalines)
        self.codestr = "VERSION"
        self.codehead = ()
        self.codetail = ("version",)
        self.coderep = ()


responses = {
    "200": LoginAcceptedResponse,
    "201": LoginAcceptedNewVerResponse,
    "203": LoggedOutResponse,
    "205": ResourceResponse,
    "206": StatsResponse,
    "207": TopResponse,
    "208": UptimeResponse,
    "209": EncryptionEnabledResponse,
    "210": MylistEntryAddedResponse,
    "211": MylistEntryDeletedResponse,
    "214": AddedFileResponse,
    "215": AddedStreamResponse,
    "219": EncodingChangedResponse,
    "220": FileResponse,
    "221": MylistResponse,
    "222": MylistStatsResponse,
    "224": NotificationResponse,
    "225": GroupstatusResponse,
    "230": AnimeResponse,
    "231": AnimeBestMatchResponse,
    "232": RandomanimeResponse,
    "240": EpisodeResponse,
    "245": ProducerResponse,
    "246": NotificationAddedResponse,
    "247": NotificationDeletedResponse,
    "248": NotificationUpdatedResponse,
    "249": MultipleNotificationResponse,
    "250": GroupResponse,
    "253": BuddyListResponse,
    "254": BuddyStateResponse,
    "255": BuddyAddedResponse,
    "256": BuddyDeletedResponse,
    "257": BuddyAcceptedResponse,
    "258": BuddyDeniedResponse,
    "260": VotedResponse,
    "261": VoteFoundResponse,
    "262": VoteUpdatedResponse,
    "263": VoteRevokedResponse,
    "270": NotificationEnabledResponse,
    "271": NotificationNotifyResponse,
    "272": NotificationMessageResponse,
    "273": NotificationBuddyResponse,
    "274": NotificationShutdownResponse,
    "280": PushackConfirmedResponse,
    "281": NotifyackSuccessfulMResponse,
    "282": NotifyackSuccessfulNResponse,
    "290": NotificationStateResponse,
    "291": NotifylistResponse,
    "292": NotifygetMessageResponse,
    "293": NotifygetNotifyResponse,
    "294": SendmsgSuccessfulResponse,
    "295": UserResponse,
    "300": PongResponse,
    "301": AuthpongResponse,
    "305": NoSuchResourceResponse,
    "309": ApiPasswordNotDefinedResponse,
    "310": FileAlreadyInMylistResponse,
    "311": MylistEntryEditedResponse,
    "312": MultipleMylistEntriesResponse,
    "314": SizeHashExistsResponse,
    "315": InvalidDataResponse,
    "316": StreamnoidUsedResponse,
    "320": NoSuchFileResponse,
    "321": NoSuchEntryResponse,
    "322": MultipleFilesFoundResponse,
    "324": NoSuchNotificationResponse,
    "325": NoGroupsFoundResponse,
    "330": NoSuchAnimeResponse,
    "340": NoSuchEpisodeResponse,
    "345": NoSuchProducerResponse,
    "350": NoSuchGroupResponse,
    "355": BuddyAlreadyAddedResponse,
    "356": NoSuchBuddyResponse,
    "357": BuddyAlreadyAcceptedResponse,
    "358": BuddyAlreadyDeniedResponse,
    "360": NoSuchVoteResponse,
    "361": InvalidVoteTypeResponse,
    "362": InvalidVoteValueResponse,
    "363": PermvoteNotAllowedResponse,
    "364": AlreadyPermvotedResponse,
    "370": NotificationDisabledResponse,
    "380": NoSuchPacketPendingResponse,
    "381": NoSuchEntryMResponse,
    "382": NoSuchEntryNResponse,
    "392": NoSuchMessageResponse,
    "393": NoSuchNotifyResponse,
    "394": NoSuchUserResponse,
    "399": NoChanges,
    "403": NotLoggedInResponse,
    "410": NoSuchMylistFileResponse,
    "411": NoSuchMylistEntryResponse,
    "500": LoginFailedResponse,
    "501": LoginFirstResponse,
    "502": AccessDeniedResponse,
    "503": ClientVersionOutdatedResponse,
    "504": ClientBannedResponse,
    "505": IllegalInputOrAccessDeniedResponse,
    "506": InvalidSessionResponse,
    "509": NoSuchEncryptionTypeResponse,
    "519": EncodingNotSupportedResponse,
    "555": BannedResponse,
    "598": UnknownCommandResponse,
    "600": InternalServerErrorResponse,
    "601": AnidbOutOfServiceResponse,
    "602": ServerBusyResponse,
    "666": ApiViolationResponse,
    "998": VersionResponse,
}



from requests.utils import dict_from_cookiejar

from sickchill import logger
from sickchill.helper.common import convert_size, try_int
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        super().__init__("SceneTime")

        self.username = None
        self.password = None
        self.minseed = 0
        self.minleech = 0

        self.enable_cookies = True

        self.cache = tvcache.TVCache(self)  # only poll SceneTime every 20 minutes max

        self.urls = {
            "base_url": "https://www.scenetime.com",
            "login": "https://www.scenetime.com/takelogin.php",
            "detail": "https://www.scenetime.com/details.php?id=%s",
            "apisearch": "https://www.scenetime.com/browse.php",
            "download": "https://www.scenetime.com/download.php/%s/%s",
        }

        self.url = self.urls["base_url"]

        self.categories = [2, 42, 9, 63, 77, 79, 100, 83]

    def login(self):
        cookie_dict = dict_from_cookiejar(self.session.cookies)
        if cookie_dict.get("uid") and cookie_dict.get("pass"):
            return True

        if self.cookies:
            success, status = self.add_cookies_from_ui()
            if not success:
                logger.info(status)
                return False

            login_params = {"username": self.username, "password": self.password}

            response = self.get_url(self.urls["login"], post_data=login_params, returns="response")
            if not response or response.status_code != 200:
                logger.warning("Unable to connect to provider")
                return False

            if dict_from_cookiejar(self.session.cookies).get("uid") in response.text:
                return True
            else:
                logger.warning("Failed to login, check your cookies")
                return False

    def search(self, search_params, age=0, ep_obj=None):
        results = []
        if not self.login():
            return results

        for mode in search_params:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))
            for search_string in search_params[mode]:

                if mode != "RSS":
                    logger.debug(_("Search String: {search_string}".format(search_string=search_string)))

                query = {"sec": "jax", "cata": "yes", "search": search_string}
                query.update({"c" + str(i): 1 for i in self.categories})

                data = self.get_url(self.urls["apisearch"], returns="text", post_data=query)

                if not data:
                    continue

                with BS4Parser(data, "html5lib") as html:
                    torrent_table = html.find(id="torrenttable")
                    if torrent_table:
                        torrent_rows = torrent_table.findAll("tr")
                    else:
                        torrent_rows = []

                    # Continue only if one Release is found
                    if len(torrent_rows) < 2:
                        logger.debug("Data returned from provider does not contain any torrents")
                        continue

                    # Scenetime apparently uses different number of cells in #torrenttable based
                    # on who you are. This works around that by extracting labels from the first
                    # <tr> and using their index to find the correct download/seeders/leechers td.
                    labels = [label.get_text(strip=True) or label.img["title"] for label in torrent_rows[0]("td")]

                    for result in torrent_rows[1:]:
                        try:
                            cells = result("td")

                            link = cells[labels.index("Name")].find("a")
                            torrent_id = link["href"].replace("details.php?id=", "").split("&")[0]

                            title = link.get_text(strip=True)
                            download_url = self.urls["download"] % (torrent_id, "{0}.torrent".format(title.replace(" ", ".")))

                            seeders = try_int(cells[labels.index("Seeders")].get_text(strip=True))
                            leechers = try_int(cells[labels.index("Leechers")].get_text(strip=True))
                            torrent_size = cells[labels.index("Size")].get_text()

                            size = convert_size(torrent_size) or -1

                        except (AttributeError, TypeError, KeyError, ValueError):
                            continue

                        if not all([title, download_url]):
                            continue

                        # Filter unseeded torrent
                        if seeders < self.minseed or leechers < self.minleech:
                            if mode != "RSS":
                                logger.debug(
                                    "Discarding torrent because it doesn't meet the minimum seeders or leechers: {0} (S:{1} L:{2})".format(
                                        title, seeders, leechers
                                    )
                                )
                            continue

                        item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": ""}
                        if mode != "RSS":
                            logger.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))

                        items.append(item)

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)

            results += items

        return results



import markdown2

from .. import logger
from ..oldbeard import helpers
from .common import PageTemplate
from .home import Home
from .routes import Route


@Route("/changes(/?.*)", name="changelog")
class HomeChangeLog(Home):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def index(self, *args, **kwargs):
        # noinspection PyBroadException
        try:
            changes = helpers.getURL("https://sickchill.github.io/sickchill-news/CHANGES.md", session=helpers.make_session(), returns="text")
        except Exception:
            logger.debug("Could not load changes from repo, giving a link!")
            changes_url = "https://sickchill.github.io/sickchill-news/CHANGES.md"
            changes = _("Could not load changes from the repo. [Click here for CHANGES.md]") + f"({changes_url})"

        t = PageTemplate(rh=self, filename="markdown.mako")
        data = markdown2.markdown(changes if changes else _("The was a problem connecting to github, please refresh and try again"), extras=["header-ids"])

        return t.render(title=_("Changelog"), header=_("Changelog"), topmenu="system", data=data, controller="changes", action="index")



from subliminal.extensions import RegistrableExtensionManager

clients = RegistrableExtensionManager(
    "oldbeard.clients",
    [
        "deluge = oldbeard.clients.deluge:Client",
        "deluged = oldbeard.clients.deluged:Client",
        "download_station = oldbeard.clients.download_station:Client",
        "mlnet = oldbeard.clients.mlnet:Client",
        "qbittorrent = oldbeard.clients.qbittorrent:Client",
        "putio = oldbeard.clients.putio:Client",
        "rtorrent = oldbeard.clients.rtorrent:Client",
        "transmission = oldbeard.clients.transmission:Client",
        "utorrent = oldbeard.clients.utorrent:Client",
    ],
)

providers = RegistrableExtensionManager(
    "oldbeard.providers",
    [
        "hdspace = oldbeard.providers.hdspace:HDSpaceProvider",
        "ncore = oldbeard.providers.ncore:NcoreProvider",
        "tntvillage = oldbeard.providers.tntvillage:TNTVillageProvider",
        "rarbg = oldbeard.providers.rarbg:RarbgProvider",
        "elitetorrent = oldbeard.providers.elitetorrent:EliteTorrentProvider",
        "hd4free = oldbeard.providers.hd4free:HD4FreeProvider",
        "horriblesubs = oldbeard.providers.horriblesubs:HorribleSubsProvider",
        "limetorrents = oldbeard.providers.limetorrents:LimeTorrentsProvider",
        "cpasbien = oldbeard.providers.cpasbien:CpasbienProvider",
        "bjshare = oldbeard.providers.bjshare:BJShareProvider",
        "tvchaosuk = oldbeard.providers.tvchaosuk:TVChaosUKProvider",
        "eztv = oldbeard.providers.eztv:EZTVProvider",
        "shazbat = oldbeard.providers.shazbat:ShazbatProvider",
        "scc = oldbeard.providers.scc:SCCProvider",
        "thepiratebay = oldbeard.providers.thepiratebay:ThePirateBayProvider",
        "torrent_paradise = oldbeard.providers.torrent_paradise:Provider",
        "nyaa = oldbeard.providers.nyaa:NyaaProvider",
        "bitcannon = oldbeard.providers.bitcannon:BitCannonProvider",
        "magnetdl = oldbeard.providers.magnetdl:MagnetDLProvider",
        "hdtorrents_it = oldbeard.providers.hdtorrents_it:HDTorrentsProvider_IT",
        "skytorrents = oldbeard.providers.skytorrents:SkyTorrents",
        "xthor = oldbeard.providers.xthor:XThorProvider",
        "norbits = oldbeard.providers.norbits:NorbitsProvider",
        "yggtorrent = oldbeard.providers.yggtorrent:YggTorrentProvider",
        "torrentbytes = oldbeard.providers.torrentbytes:TorrentBytesProvider",
        "btn = oldbeard.providers.btn:BTNProvider",
        "demonoid = oldbeard.providers.demonoid:DemonoidProvider",
        "tokyotoshokan = oldbeard.providers.tokyotoshokan:TokyoToshokanProvider",
        "speedcd = oldbeard.providers.speedcd:SpeedCDProvider",
        "kat = oldbeard.providers.kat:KatProvider",
        "scenetime = oldbeard.providers.scenetime:SceneTimeProvider",
        "alpharatio = oldbeard.providers.alpharatio:AlphaRatioProvider",
        "hounddawgs = oldbeard.providers.hounddawgs:HoundDawgsProvider",
        "hdbits = oldbeard.providers.hdbits:HDBitsProvider",
        "pretome = oldbeard.providers.pretome:PretomeProvider",
        "binsearch = oldbeard.providers.binsearch:BinSearchProvider",
        "morethantv = oldbeard.providers.morethantv:MoreThanTVProvider",
        "torrentz = oldbeard.providers.torrentz:TorrentzProvider",
        "archetorrent = oldbeard.providers.archetorrent:ArcheTorrentProvider",
        "hdtorrents = oldbeard.providers.hdtorrents:HDTorrentsProvider",
        "immortalseed = oldbeard.providers.immortalseed:ImmortalseedProvider",
        "ilcorsaronero = oldbeard.providers.ilcorsaronero:ilCorsaroNeroProvider",
        "iptorrents = oldbeard.providers.iptorrents:IPTorrentsProvider",
        "torrent9 = oldbeard.providers.torrent9:Torrent9Provider",
        "newpct = oldbeard.providers.newpct:newpctProvider",
        "filelist = oldbeard.providers.filelist:FileListProvider",
        "abnormal = oldbeard.providers.abnormal:ABNormalProvider",
        "torrentproject = oldbeard.providers.torrentproject:TorrentProjectProvider",
        "danishbits = oldbeard.providers.danishbits:DanishbitsProvider",
        "omgwtfnzbs = oldbeard.providers.omgwtfnzbs:OmgwtfnzbsProvider",
        "gimmepeers = oldbeard.providers.gimmepeers:GimmePeersProvider",
        "nebulance = oldbeard.providers.nebulance:NebulanceProvider",
        "torrentday = oldbeard.providers.torrentday:TorrentDayProvider",
        "torrentleech = oldbeard.providers.torrentleech:TorrentLeechProvider",
    ],
)

notifiers = RegistrableExtensionManager(
    "oldbeard.notifiers",
    [
        "plex = oldbeard.notifiers.plex:Notifier",
        "synoindex = oldbeard.notifiers.synoindex:Notifier",
        "telegram = oldbeard.notifiers.telegram:Notifier",
        "trakt = oldbeard.notifiers.trakt:Notifier",
        "join = oldbeard.notifiers.join:Notifier",
        "growl = oldbeard.notifiers.growl:Notifier",
        "emby = oldbeard.notifiers.emby:Notifier",
        "pushbullet = oldbeard.notifiers.pushbullet:Notifier",
        "nmjv2 = oldbeard.notifiers.nmjv2:Notifier",
        "tweet = oldbeard.notifiers.tweet:Notifier",
        "growl = oldbeard.notifiers.growl:Notifier",
        "pushalot = oldbeard.notifiers.pushalot:Notifier",
        "emailnotify = oldbeard.notifiers.emailnotify:Notifier",
        "prowl = oldbeard.notifiers.prowl:Notifier",
        "boxcar2 = oldbeard.notifiers.boxcar2:Notifier",
        "pytivo = oldbeard.notifiers.pytivo:Notifier",
        "boxcar2 = oldbeard.notifiers.boxcar2:Notifier",
        "nmj = oldbeard.notifiers.nmj:Notifier",
        "synologynotifier = oldbeard.notifiers.synologynotifier:Notifier",
        "matrix = oldbeard.notifiers.matrix:Notifier",
        "rocketchat = oldbeard.notifiers.rocketchat:Notifier",
        "trakt = oldbeard.notifiers.trakt:Notifier",
        "libnotify = oldbeard.notifiers.libnotify:Notifier",
        "rocketchat = oldbeard.notifiers.rocketchat:Notifier",
        "discord = oldbeard.notifiers.discord:Notifier",
        "pushbullet = oldbeard.notifiers.pushbullet:Notifier",
        # "twilio_notify = oldbeard.notifiers.twilio_notify:Notifier",
        "tweet = oldbeard.notifiers.tweet:Notifier",
        "synologynotifier = oldbeard.notifiers.synologynotifier:Notifier",
        "plex = oldbeard.notifiers.plex:Notifier",
        "discord = oldbeard.notifiers.discord:Notifier",
        "freemobile = oldbeard.notifiers.freemobile:Notifier",
        "slack = oldbeard.notifiers.slack:Notifier",
        "matrix = oldbeard.notifiers.matrix:Notifier",
        "synoindex = oldbeard.notifiers.synoindex:Notifier",
        "kodi = oldbeard.notifiers.kodi:Notifier",
        "nmjv2 = oldbeard.notifiers.nmjv2:Notifier",
        "emailnotify = oldbeard.notifiers.emailnotify:Notifier",
        "kodi = oldbeard.notifiers.kodi:Notifier",
        "pushalot = oldbeard.notifiers.pushalot:Notifier",
        "prowl = oldbeard.notifiers.prowl:Notifier",
        "pushover = oldbeard.notifiers.pushover:Notifier",
        "telegram = oldbeard.notifiers.telegram:Notifier",
        "pytivo = oldbeard.notifiers.pytivo:Notifier",
        "join = oldbeard.notifiers.join:Notifier",
        "pushover = oldbeard.notifiers.pushover:Notifier",
        "base = oldbeard.notifiers.base:Notifier",
        "emby = oldbeard.notifiers.emby:Notifier",
        "slack = oldbeard.notifiers.slack:Notifier",
        "freemobile = oldbeard.notifiers.freemobile:Notifier",
        "libnotify = oldbeard.notifiers.libnotify:Notifier",
        "nmj = oldbeard.notifiers.nmj:Notifier",
    ],
)

metadata = RegistrableExtensionManager(
    "sickchill.providers.metadata",
    [
        "mede8er = sickchill.providers.metadata.mede8er:Mede8erMetadata",
        "ps3 = sickchill.providers.metadata.ps3:PS3Metadata",
        "tivo = sickchill.providers.metadata.tivo:TIVOMetadata",
        "mede8er = sickchill.providers.metadata.mede8er:Mede8erMetadata",
        "tivo = sickchill.providers.metadata.tivo:TIVOMetadata",
        "mediabrowser = sickchill.providers.metadata.mediabrowser:MediaBrowserMetadata",
        "wdtv = sickchill.providers.metadata.wdtv:WDTVMetadata",
        "kodi = sickchill.providers.metadata.kodi:KODIMetadata",
        "wdtv = sickchill.providers.metadata.wdtv:WDTVMetadata",
        "kodi = sickchill.providers.metadata.kodi:KODIMetadata",
        "mediabrowser = sickchill.providers.metadata.mediabrowser:MediaBrowserMetadata",
        "ps3 = sickchill.providers.metadata.ps3:PS3Metadata",
    ],
)



import json
import time
import traceback

import requests

from sickchill import logger, settings

from . import common, generic_queue

DISCORD = 600


class NotificationsQueue(generic_queue.GenericQueue):
    """
    Queue to handle multiple post processing tasks
    """

    def __init__(self):
        """

        :rtype: object
        """
        super().__init__()
        self.queue_name = "NOTIFICATIONS"

    @property
    def is_paused(self):
        """
        Shows if the post processing queue is paused
        :return: bool
        """
        return self.min_priority == generic_queue.QueuePriorities.HIGH

    @property
    def pause(self):
        """
        Pause the post processing queue
        """
        self.min_priority = generic_queue.QueuePriorities.HIGH
        return True

    @property
    def unpause(self):
        """
        Unpause the processing queue
        """
        self.min_priority = 0
        return True

    def __len__(self):
        size = len(self.queue)
        if self.currentItem:
            size += 1
        return size

    def add_item(self, message, notifier="discord", force_next=False):
        added = False
        item = None
        with self.lock:
            if self.queue and not force_next:
                for index in range(0, len(self.queue)):
                    if isinstance(self.queue[index], DiscordTask):
                        if len(self.queue[index]) < 24:
                            self.queue[index].append(message)
                            added = True
                            break
            if not added:
                item = DiscordTask(message)
                if force_next:
                    item.run()
                    return item.last_result
        if not added:
            added = super().add_item(item)
        return added


class DiscordTask(generic_queue.QueueItem):
    """
    Processing task
    """

    def __init__(self, message):
        super().__init__("Discord", DISCORD)

        self.embed = {
            "author": {
                "name": "SickChill",
                # 'url':
                "icon_url": settings.DISCORD_AVATAR_URL,
            },
            "fields": [],
        }

        self.priority = generic_queue.QueuePriorities.LOW
        self.last_result = None

        self.append(message)

    def run(self):
        """
        Runs the task
        :return: None
        """
        super().run()

        # noinspection PyBroadException
        try:
            logger.debug("Task for {} started".format(self.action_id))
            self.last_result = self._send_discord()
            logger.debug("Task for {} completed".format(self.action_id))

            # give the CPU a break
            time.sleep(common.cpu_presets[settings.CPU_PRESET])
        except Exception:
            logger.debug(traceback.format_exc())

        super().finish()
        self.finish()

    def append(self, message):
        self.embed["fields"] += [{"name": "Notification", "value": message}]

    def __len__(self):
        return len(self.embed["fields"])

    def _send_discord(self, webhook: str = None, name: str = None, avatar: str = None, tts=None):
        discord_webhook = webhook or settings.DISCORD_WEBHOOK
        discord_name = name or settings.DISCORD_NAME
        avatar_icon = avatar or settings.DISCORD_AVATAR_URL
        discord_tts = bool(settings.DISCORD_TTS if tts is None else tts)

        logger.info("Sending discord message: " + ", ".join(f["value"] for f in self.embed["fields"]))
        logger.info("Sending discord message to url: " + discord_webhook)

        headers = {"Content-Type": "application/json"}
        try:
            r = requests.post(
                discord_webhook, data=json.dumps(dict(embeds=[self.embed], username=discord_name, avatar_url=avatar_icon, tts=discord_tts)), headers=headers
            )
            r.raise_for_status()
        except requests.exceptions.ConnectionError as error:
            logger.info("Could not reach the webhook url")
            return False
        except requests.exceptions.RequestException as error:
            if error.response.status_code != 429 or int(error.response.headers.get("X-RateLimit-Remaining")) != 0:
                raise error

            logger.info("Discord rate limiting, retrying after {} seconds".format(error.response.headers.get("X-RateLimit-Reset-After")))
            time.sleep(int(error.response.headers.get("X-RateLimit-Reset-After")) + 1)
            r = requests.post(
                discord_webhook, data=json.dumps(dict(embeds=[self.embed], username=discord_name, avatar_url=avatar_icon, tts=discord_tts)), headers=headers
            )
            r.raise_for_status()
        except Exception as error:
            logger.exception("Error Sending Discord message: " + str(error))

            return False

        return True



import re
import time
import traceback

from sickchill import logger, settings
from sickchill.helper.common import try_int
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.oldbeard.common import cpu_presets
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        super().__init__("EliteTorrent")

        self.onlyspasearch = None
        self.minseed = 0
        self.minleech = 0
        self.cache = tvcache.TVCache(self)  # Only poll EliteTorrent every 20 minutes max

        self.urls = {"base_url": "https://www.elitetorrent.eu", "search": "https://www.elitetorrent.eu/torrents.php"}

        self.url = self.urls["base_url"]

    def search(self, search_strings, age=0, ep_obj=None):
        results = []
        lang_info = "" if not ep_obj or not ep_obj.show else ep_obj.show.lang

        """
        Search query:
        https://www.elitetorrent.eu/torrents.php?cat=4&modo=listado&orden=fecha&pag=1&buscar=fringe

        cat = 4 => Shows
        modo = listado => display results mode
        orden = fecha => order
        buscar => Search show
        pag = 1 => page number
        """

        search_params = {"cat": 4, "modo": "listado", "orden": "fecha", "pag": 1, "buscar": ""}

        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))

            # Only search if user conditions are true
            if self.onlyspasearch and lang_info != "es" and mode != "RSS":
                logger.debug("Show info is not spanish, skipping provider search")
                continue

            for search_string in {*search_strings[mode]}:
                if mode != "RSS":
                    logger.debug("Search string: {0}".format(search_string))

                search_string = re.sub(r"S0*(\d*)E(\d*)", r"\1x\2", search_string)
                search_params["buscar"] = search_string.strip() if mode != "RSS" else ""

                time.sleep(cpu_presets[settings.CPU_PRESET])
                data = self.get_url(self.urls["search"], params=search_params, returns="text")
                if not data:
                    continue

                try:
                    with BS4Parser(data, "html5lib") as html:
                        torrent_table = html.find("table", class_="fichas-listado")
                        torrent_rows = torrent_table("tr") if torrent_table else []

                        if len(torrent_rows) < 2:
                            logger.debug("Data returned from provider does not contain any torrents")
                            continue

                        for row in torrent_rows[1:]:
                            try:
                                download_url = self.urls["base_url"] + row.find("a")["href"]

                                """
                                Transform from
                                https://elitetorrent.eu/torrent/40142/la-zona-1x02
                                to
                                https://elitetorrent.eu/get-torrent/40142
                                """

                                download_url = re.sub(r"/torrent/(\d*)/.*", r"/get-torrent/\1", download_url)

                                """
                                Trick for accents for this provider.

                                - data = self.get_url(self.urls['search'], params=search_params, returns='text') -
                                returns latin1 coded text and this makes that the title used for the search
                                and the title retrieved from the parsed web page doesn't match so I get
                                "No needed episodes found during backlog search for: XXXX"

                                This is not the best solution but it works.

                                First encode latin1 and then decode utf8 to remains str
                                """
                                row_title = row.find("a", class_="nombre")["title"]
                                title = self._processTitle(row_title.encode("latin-1").decode("utf8"))

                                seeders = try_int(row.find("td", class_="semillas").get_text(strip=True))
                                leechers = try_int(row.find("td", class_="clientes").get_text(strip=True))

                                # seeders are not well reported. Set 1 in case of 0
                                seeders = max(1, seeders)

                                # Provider does not provide size
                                size = -1

                            except (AttributeError, TypeError, KeyError, ValueError):
                                continue

                            if not all([title, download_url]):
                                continue

                            # Filter unseeded torrent
                            if seeders < self.minseed or leechers < self.minleech:
                                if mode != "RSS":
                                    logger.debug(
                                        "Discarding torrent because it doesn't meet the minimum seeders or leechers: {0} (S:{1} L:{2})".format(
                                            title, seeders, leechers
                                        )
                                    )
                                continue

                            item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": ""}
                            if mode != "RSS":
                                logger.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))

                            items.append(item)

                except Exception:
                    logger.warning("Failed parsing provider. Traceback: {0}".format(traceback.format_exc()))

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)

            results += items

        return results

    @staticmethod
    def _processTitle(title):

        # Quality, if no literal is defined it's HDTV
        if "calidad" not in title:
            title += " HDTV x264"

        title = title.replace("(calidad baja)", "HDTV x264")
        title = title.replace("(Buena calidad)", "720p HDTV x264")
        title = title.replace("(Alta calidad)", "720p HDTV x264")
        title = title.replace("(calidad regular)", "DVDrip x264")
        title = title.replace("(calidad media)", "DVDrip x264")

        # Language, all results from this provider have spanish audio, we append it to title (avoid to download undesired torrents)
        title += " SPANISH AUDIO"
        title += "-ELITETORRENT"

        return title.strip()



# Author: Rafael Silva <rpluto@gmail.com>
# Author: Marvin Pinto <me@marvinp.ca>
from sickchill import logger, settings
from sickchill.oldbeard import common, helpers


class Notifier(object):
    def __init__(self):
        self.session = helpers.make_session()
        self.url = "https://new.boxcar.io/api/notifications"

    def test_notify(self, accesstoken, title="SickChill : Test"):
        return self._sendBoxcar2("This is a test notification from SickChill", title, accesstoken)

    def _sendBoxcar2(self, msg, title, accesstoken):
        """
        Sends a boxcar2 notification to the address provided

        msg: The message to send
        title: The title of the message
        accesstoken: to send to this device

        returns: True if the message succeeded, False otherwise
        """
        # http://blog.boxcar.io/post/93211745502/boxcar-api-update-boxcar-api-update-icon-and

        post_data = {
            "user_credentials": accesstoken,
            "notification[title]": "SickChill : {0}: {1}".format(title, msg),
            "notification[long_message]": msg,
            "notification[sound]": "notifier-2",
            "notification[source_name]": "SickChill",
            "notification[icon_url]": settings.LOGO_URL,
        }

        response = helpers.getURL(self.url, post_data=post_data, session=self.session, timeout=60, returns="json")
        if not response:
            logger.exception("Boxcar2 notification failed.")
            return False

        logger.debug("Boxcar2 notification successful.")
        return True

    def notify_snatch(self, ep_name, title=common.notifyStrings[common.NOTIFY_SNATCH]):
        if settings.BOXCAR2_NOTIFY_ONSNATCH:
            self._notifyBoxcar2(title, ep_name)

    def notify_download(self, ep_name, title=common.notifyStrings[common.NOTIFY_DOWNLOAD]):
        if settings.BOXCAR2_NOTIFY_ONDOWNLOAD:
            self._notifyBoxcar2(title, ep_name)

    def notify_subtitle_download(self, ep_name, lang, title=common.notifyStrings[common.NOTIFY_SUBTITLE_DOWNLOAD]):
        if settings.BOXCAR2_NOTIFY_ONSUBTITLEDOWNLOAD:
            self._notifyBoxcar2(title, ep_name + ": " + lang)

    def notify_git_update(self, new_version="??"):
        update_text = common.notifyStrings[common.NOTIFY_GIT_UPDATE_TEXT]
        title = common.notifyStrings[common.NOTIFY_GIT_UPDATE]
        self._notifyBoxcar2(title, update_text + new_version)

    def notify_login(self, ipaddress=""):
        update_text = common.notifyStrings[common.NOTIFY_LOGIN_TEXT]
        title = common.notifyStrings[common.NOTIFY_LOGIN]
        self._notifyBoxcar2(title, update_text.format(ipaddress))

    def _notifyBoxcar2(self, title, message, accesstoken=None):
        """
        Sends a boxcar2 notification based on the provided info or SB config

        title: The title of the notification to send
        message: The message string to send
        accesstoken: to send to this device
        """

        if not settings.USE_BOXCAR2:
            logger.debug("Notification for Boxcar2 not enabled, skipping this notification")
            return False

        accesstoken = accesstoken or settings.BOXCAR2_ACCESSTOKEN

        logger.debug("Sending notification for {0}".format(message))

        return self._sendBoxcar2(message, title, accesstoken)



import re
from urllib.parse import quote_plus

from requests.utils import dict_from_cookiejar

from sickchill import logger
from sickchill.helper.common import convert_size, try_int
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        super().__init__("HDTorrents")

        self.username = None
        self.password = None
        self.minseed = 0
        self.minleech = 0
        self.freeleech = None

        self.urls = {
            "base_url": "https://hd-torrents.org",
            "login": "https://hd-torrents.org/login.php",
            "search": "https://hd-torrents.org/torrents.php?search=%s&active=1&options=0%s",
            "rss": "https://hd-torrents.org/torrents.php?search=&active=1&options=0%s",
            "home": "https://hd-torrents.org/%s",
        }

        self.url = self.urls["base_url"]

        self.categories = "&category[]=59&category[]=60&category[]=30&category[]=38&category[]=65"
        self.proper_strings = ["PROPER", "REPACK"]

        self.cache = tvcache.TVCache(self, min_time=30)  # only poll HDTorrents every 30 minutes max

    def _check_auth(self):

        if not self.username or not self.password:
            logger.warning("Invalid username or password. Check your settings")

        return True

    def login(self):
        if any(dict_from_cookiejar(self.session.cookies).values()):
            return True

        login_params = {"uid": self.username, "pwd": self.password, "submit": "Confirm"}

        response = self.get_url(self.urls["login"], post_data=login_params, returns="text")
        if not response:
            logger.warning("Unable to connect to provider")
            return False

        if re.search("You need cookies enabled to log in.", response):
            logger.warning("Invalid username or password. Check your settings")
            return False

        return True

    def search(self, search_strings, age=0, ep_obj=None):
        results = []
        if not self.login():
            return results

        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))
            for search_string in {*search_strings[mode]}:

                if mode != "RSS":
                    search_url = self.urls["search"] % (quote_plus(search_string), self.categories)
                    logger.debug(_("Search String: {search_string}".format(search_string=search_string)))
                else:
                    search_url = self.urls["rss"] % self.categories

                if self.freeleech:
                    search_url = search_url.replace("active=1", "active=5")

                data = self.get_url(search_url, returns="text")
                if not data or "please try later" in data:
                    logger.debug("No data returned from provider")
                    continue

                if data.find("No torrents here") != -1:
                    logger.debug("Data returned from provider does not contain any torrents")
                    continue

                # Search result page contains some invalid html that prevents html parser from returning all data.
                # We cut everything before the table that contains the data we are interested in thus eliminating
                # the invalid html portions
                try:
                    index = data.lower().index('<table class="mainblockcontenttt"')
                except ValueError:
                    logger.debug("Could not find table of torrents mainblockcontenttt")
                    continue

                data = data[index:]

                with BS4Parser(data, "html5lib") as html:
                    if not html:
                        logger.debug("No html data parsed from provider")
                        continue

                    torrent_rows = []
                    torrent_table = html.find("table", class_="mainblockcontenttt")
                    if torrent_table:
                        torrent_rows = torrent_table("tr")

                    if not torrent_rows:
                        logger.debug("Could not find results in returned data")
                        continue

                    # Cat., Active, Filename, Dl, Wl, Added, Size, Uploader, S, L, C
                    labels = [label.a.get_text(strip=True) if label.a else label.get_text(strip=True) for label in torrent_rows[0]("td")]

                    # Skip column headers
                    for result in torrent_rows[1:]:
                        try:
                            cells = result.find_all("td")[: len(labels)]
                            if len(cells) < len(labels):
                                continue

                            title = cells[labels.index("Filename")].a.get_text(strip=True)
                            seeders = try_int(cells[labels.index("S")].get_text(strip=True))
                            leechers = try_int(cells[labels.index("L")].get_text(strip=True))
                            torrent_size = cells[labels.index("Size")].get_text()

                            size = convert_size(torrent_size) or -1
                            download_url = self.url + "/" + cells[labels.index("Dl")].a["href"]
                        except (AttributeError, TypeError, KeyError, ValueError, IndexError):
                            continue

                        if not all([title, download_url]):
                            continue

                        # Filter unseeded torrent
                        if seeders < self.minseed or leechers < self.minleech:
                            if mode != "RSS":
                                logger.debug(
                                    "Discarding torrent because it doesn't meet the minimum seeders or leechers: {0} (S:{1} L:{2})".format(
                                        title, seeders, leechers
                                    )
                                )
                            continue

                        item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": ""}
                        if mode != "RSS":
                            logger.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))

                        items.append(item)

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)

            results += items

        return results



import traceback
from functools import wraps

import requests
from urllib3.exceptions import HTTPError, HTTPWarning

from sickchill import logger


class ExceptionDecorator(object):
    def __init__(
        self,
        default_return=None,
        catch=(HTTPError, HTTPWarning, requests.exceptions.RequestException, requests.exceptions.RequestsWarning, TypeError),
        image_api=False,
    ):
        self.default_return = default_return or []
        self.catch = catch
        self.image_api = image_api

    def __call__(self, target, *args, **kwargs):
        @wraps(target)
        def wrapper(*args, **kwargs):
            try:
                result = target(*args, **kwargs)
            except self.catch as error:
                if self.image_api and not kwargs.get("lang"):
                    if args[1].lang != "en":
                        logger.debug("Could not find the image on the indexer, re-trying to find it in english")
                        kwargs["lang"] = "en"
                        return wrapper(*args, **kwargs)

                logger.debug(f"Could not find item on the indexer: (Indexer probably does not have this item) [{error}]")
                logger.debug(traceback.format_exc())
                result = self.default_return
            except requests.exceptions.RequestException as error:
                logger.debug(f"Could not find item on the indexer: (Indexer probably doesn't have this item) [{error}]")
                logger.debug(traceback.format_exc())
                result = self.default_return

            return result

        wrapper.__doc__ = target.__doc__
        return wrapper



from urllib.parse import urljoin

from sickchill import logger
from sickchill.helper.common import convert_size, try_int
from sickchill.oldbeard import tvcache
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        super().__init__("HD4Free")

        self.url = "https://hd4free.xyz"
        self.urls = {"search": urljoin(self.url, "/searchapi.php")}

        self.freeleech = None
        self.username = None
        self.api_key = None
        self.minseed = 0
        self.minleech = 0

        self.cache = tvcache.TVCache(self, min_time=10)  # Only poll HD4Free every 10 minutes max

    def _check_auth(self):
        if self.username and self.api_key:
            return True

        logger.warning("Your authentication credentials for {0} are missing, check your config.".format(self.name))
        return False

    def search(self, search_strings, age=0, ep_obj=None):
        results = []
        if not self._check_auth:
            return results

        search_params = {"tv": "true", "username": self.username, "apikey": self.api_key}

        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))
            for search_string in {*search_strings[mode]}:
                if self.freeleech:
                    search_params["fl"] = "true"
                else:
                    search_params.pop("fl", "")

                if mode != "RSS":
                    logger.debug("Search string: " + search_string.strip())
                    search_params["search"] = search_string
                else:
                    search_params.pop("search", "")

                try:
                    jdata = self.get_url(self.urls["search"], params=search_params, returns="json")
                except ValueError:
                    logger.debug("No data returned from provider")
                    continue

                if not jdata:
                    logger.debug("No data returned from provider")
                    continue

                error = jdata.get("error")
                if error:
                    logger.debug("{}".format(error))
                    return results

                try:
                    if jdata["0"]["total_results"] == 0:
                        logger.debug("Provider has no results for this search")
                        continue
                except Exception:
                    continue

                for i in jdata:
                    try:
                        title = jdata[i]["release_name"]
                        download_url = jdata[i]["download_url"]
                        if not all([title, download_url]):
                            continue

                        seeders = jdata[i]["seeders"]
                        leechers = jdata[i]["leechers"]
                        if seeders < self.minseed or leechers < self.minleech:
                            if mode != "RSS":
                                logger.debug(
                                    "Discarding torrent because it doesn't meet the minimum seeders or leechers: {0} (S:{1} L:{2})".format(
                                        title, seeders, leechers
                                    )
                                )
                            continue

                        torrent_size = str(jdata[i]["size"]) + " MB"
                        size = convert_size(torrent_size) or -1
                        item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": ""}

                        if mode != "RSS":
                            logger.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))

                        items.append(item)
                    except Exception:
                        continue

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)

            results += items

        return results



import bisect
import io
import json
import logging
import zipfile
from collections import defaultdict

from babelfish import Language
from guessit import guessit
from requests import Session
from subliminal import __short_version__
from subliminal.cache import region, SHOW_EXPIRATION_TIME
from subliminal.exceptions import AuthenticationError, ConfigurationError, ProviderError
from subliminal.matches import guess_matches
from subliminal.providers import ParserBeautifulSoup, Provider
from subliminal.subtitle import fix_line_ending, Subtitle
from subliminal.utils import sanitize
from subliminal.video import Episode, Movie

logger = logging.getLogger(__name__)


class SubsCenterSubtitle(Subtitle):
    """SubsCenter Subtitle."""

    provider_name = "subscenter"

    def __init__(
        self, language, hearing_impaired, page_link, series, season, episode, title, subtitle_id, subtitle_key, subtitle_version, downloaded, releases
    ):
        super().__init__(language, hearing_impaired, page_link)
        self.series = series
        self.season = season
        self.episode = episode
        self.title = title
        self.subtitle_id = subtitle_id
        self.subtitle_key = subtitle_key
        self.subtitle_version = subtitle_version
        self.downloaded = downloaded
        self.releases = releases

    @property
    def id(self):
        return str(self.subtitle_id)

    def get_matches(self, video):
        matches = set()

        # episode
        if isinstance(video, Episode):
            # series
            if video.series and sanitize(self.series) == sanitize(video.series):
                matches.add("series")
            # season
            if video.season and self.season == video.season:
                matches.add("season")
            # episode
            if video.episode and self.episode == video.episode:
                matches.add("episode")
            # guess
            for release in self.releases:
                matches |= guess_matches(video, guessit(release, {"type": "episode"}))
        # movie
        elif isinstance(video, Movie):
            # guess
            for release in self.releases:
                matches |= guess_matches(video, guessit(release, {"type": "movie"}))

        # title
        if video.title and sanitize(self.title) == sanitize(video.title):
            matches.add("title")

        return matches


class SubsCenterProvider(Provider):
    """SubsCenter Provider."""

    languages = {Language.fromalpha2(l) for l in ["he"]}
    server_url = "http://www.subscenter.info/he/"

    def __init__(self, username=None, password=None):
        if username is not None and password is None or username is None and password is not None:
            raise ConfigurationError("Username and password must be specified")

        self.session = None
        self.username = username
        self.password = password
        self.logged_in = False

    def initialize(self):
        self.session = Session()
        self.session.headers["User-Agent"] = "Subliminal/{}".format(__short_version__)

        # login
        if self.username is not None and self.password is not None:
            logger.debug("Logging in")
            url = self.server_url + "subscenter/accounts/login/"

            # retrieve CSRF token
            self.session.get(url)
            csrf_token = self.session.cookies["csrftoken"]

            # actual login
            data = {"username": self.username, "password": self.password, "csrfmiddlewaretoken": csrf_token}
            r = self.session.post(url, data, allow_redirects=False, timeout=10)

            if r.status_code != 302:
                raise AuthenticationError(self.username)

            logger.info("Logged in")
            self.logged_in = True

    def terminate(self):
        # logout
        if self.logged_in:
            logger.info("Logging out")
            r = self.session.get(self.server_url + "subscenter/accounts/logout/", timeout=10)
            r.raise_for_status()
            logger.info("Logged out")
            self.logged_in = False

        self.session.close()

    @region.cache_on_arguments(expiration_time=SHOW_EXPIRATION_TIME)
    def _search_url_titles(self, title):
        """Search the URL titles by kind for the given `title`.

        :param str title: title to search for.
        :return: the URL titles by kind.
        :rtype: collections.defaultdict

        """
        # make the search
        logger.info("Searching title name for %r", title)
        r = self.session.get(self.server_url + "subtitle/search/", params={"q": title}, timeout=10)
        r.raise_for_status()

        # check for redirections
        if r.history and all([h.status_code == 302 for h in r.history]):
            logger.debug("Redirected to the subtitles page")
            links = [r.url]
        else:
            # get the suggestions (if needed)
            soup = ParserBeautifulSoup(r.content, ["lxml", "html.parser"])
            links = [link.attrs["href"] for link in soup.select("#processes div.generalWindowTop a")]
            logger.debug("Found %d suggestions", len(links))

        url_titles = defaultdict(list)
        for link in links:
            parts = link.split("/")
            url_titles[parts[-3]].append(parts[-2])

        return url_titles

    def query(self, title, season=None, episode=None):
        # search for the url title
        url_titles = self._search_url_titles(title)

        # episode
        if season and episode:
            if "series" not in url_titles:
                logger.error("No URL title found for series %r", title)
                return []
            url_title = url_titles["series"][0]
            logger.debug("Using series title %r", url_title)
            url = self.server_url + "cst/data/series/sb/{}/{}/{}/".format(url_title, season, episode)
            page_link = self.server_url + "subtitle/series/{}/{}/{}/".format(url_title, season, episode)
        else:
            if "movie" not in url_titles:
                logger.error("No URL title found for movie %r", title)
                return []
            url_title = url_titles["movie"][0]
            logger.debug("Using movie title %r", url_title)
            url = self.server_url + "cst/data/movie/sb/{}/".format(url_title)
            page_link = self.server_url + "subtitle/movie/{}/".format(url_title)

        # get the list of subtitles
        logger.debug("Getting the list of subtitles")
        r = self.session.get(url)
        r.raise_for_status()
        results = json.loads(r.text)

        # loop over results
        subtitles = {}
        for language_code, language_data in results.items():
            for quality_data in language_data.values():
                for quality, subtitles_data in quality_data.items():
                    for subtitle_item in subtitles_data.values():
                        # read the item
                        language = Language.fromalpha2(language_code)
                        hearing_impaired = bool(subtitle_item["hearing_impaired"])
                        subtitle_id = subtitle_item["id"]
                        subtitle_key = subtitle_item["key"]
                        subtitle_version = subtitle_item["h_version"]
                        downloaded = subtitle_item["downloaded"]
                        release = subtitle_item["subtitle_version"]

                        # add the release and increment downloaded count if we already have the subtitle
                        if subtitle_id in subtitles:
                            logger.debug("Found additional release %r for subtitle %d", release, subtitle_id)
                            bisect.insort_left(subtitles[subtitle_id].releases, release)  # deterministic order
                            subtitles[subtitle_id].downloaded += downloaded
                            continue

                        # otherwise create it
                        subtitle = SubsCenterSubtitle(
                            language,
                            hearing_impaired,
                            page_link,
                            title,
                            season,
                            episode,
                            title,
                            subtitle_id,
                            subtitle_key,
                            subtitle_version,
                            downloaded,
                            [release],
                        )
                        logger.debug("Found subtitle %r", subtitle)
                        subtitles[subtitle_id] = subtitle

        return list(subtitles.values())

    def list_subtitles(self, video: Episode, languages):
        season = episode = None
        title = video.title

        if isinstance(video, Episode):
            title = video.series
            season = video.season
            episode = video.episode

        return [s for s in self.query(title, season, episode) if s.language in languages]

    def download_subtitle(self, subtitle: SubsCenterSubtitle):
        # download
        url = self.server_url + "subtitle/download/{}/{}/".format(subtitle.language.alpha2, subtitle.subtitle_id)
        params = {"v": subtitle.subtitle_version, "key": subtitle.subtitle_key}
        r = self.session.get(url, params=params, headers={"Referer": subtitle.page_link}, timeout=10)
        r.raise_for_status()

        # open the zip
        try:
            with zipfile.ZipFile(io.BytesIO(r.content)) as zf:
                # remove some filenames from the namelist
                namelist = [n for n in zf.namelist() if not n.endswith(".txt")]
                if len(namelist) > 1:
                    raise ProviderError("More than one file to unzip")

                subtitle.content = fix_line_ending(zf.read(namelist[0]))
        except zipfile.BadZipfile:
            # if no zip file was retrieved, daily downloads limit has exceeded
            raise ProviderError("Daily limit exceeded")



import re
import traceback

from requests.utils import dict_from_cookiejar

from sickchill import logger
from sickchill.helper.common import convert_size, try_int
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        super().__init__("HoundDawgs")

        self.username = None
        self.password = None
        self.minseed = 0
        self.minleech = 0
        self.freeleech = None
        self.ranked = None

        self.urls = {"base_url": "https://hounddawgs.org/", "search": "https://hounddawgs.org/torrents.php", "login": "https://hounddawgs.org/login.php"}

        self.url = self.urls["base_url"]

        self.search_params = {
            "filter_cat[85]": 1,
            "filter_cat[58]": 1,
            "filter_cat[57]": 1,
            "filter_cat[74]": 1,
            "filter_cat[92]": 1,
            "filter_cat[93]": 1,
            "order_by": "s3",
            "order_way": "desc",
            "type": "",
            "userid": "",
            "searchstr": "",
            "searchimdb": "",
            "searchtags": "",
        }

        self.cache = tvcache.TVCache(self)

    def login(self):
        if any(dict_from_cookiejar(self.session.cookies).values()):
            return True

        login_params = {"username": self.username, "password": self.password, "keeplogged": "on", "login": "Login"}

        self.get_url(self.urls["base_url"], returns="text")
        response = self.get_url(self.urls["login"], post_data=login_params, returns="text")
        if not response:
            logger.warning("Unable to connect to provider")
            return False

        if (
            re.search("Dit brugernavn eller kodeord er forkert.", response)
            or re.search("<title>Login :: HoundDawgs</title>", response)
            or re.search("Dine cookies er ikke aktiveret.", response)
        ):
            logger.warning("Invalid username or password. Check your settings")
            return False

        return True

    def search(self, search_strings, age=0, ep_obj=None):
        results = []
        if not self.login():
            return results

        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))
            for search_string in {*search_strings[mode]}:

                if mode != "RSS":
                    logger.debug(_("Search String: {search_string}".format(search_string=search_string)))

                self.search_params["searchstr"] = search_string

                data = self.get_url(self.urls["search"], params=self.search_params, returns="text")
                if not data:
                    logger.debug("URL did not return data")
                    continue

                strTableStart = '<table class="torrent_table'
                startTableIndex = data.find(strTableStart)
                trimmedData = data[startTableIndex:]
                if not trimmedData:
                    continue

                try:
                    with BS4Parser(trimmedData, "html5lib") as html:
                        result_table = html.find("table", {"id": "torrent_table"})

                        if not result_table:
                            logger.debug("Data returned from provider does not contain any torrents")
                            continue

                        result_tbody = result_table.find("tbody")
                        entries = result_tbody.contents
                        del entries[1::2]

                        for result in entries[1:]:

                            torrent = result("td")
                            if len(torrent) <= 1:
                                break

                            allAs = (torrent[1])("a")

                            try:
                                notinternal = result.find("img", src="/static//common/user_upload.png")
                                if self.ranked and notinternal:
                                    logger.debug("Found a user uploaded release, Ignoring it..")
                                    continue
                                freeleech = result.find("img", src="/static//common/browse/freeleech.png")
                                if self.freeleech and not freeleech:
                                    continue
                                title = allAs[2].string
                                download_url = self.urls["base_url"] + allAs[0].attrs["href"]
                                torrent_size = result.find("td", class_="nobr").find_next_sibling("td").string
                                if torrent_size:
                                    size = convert_size(torrent_size) or -1
                                seeders = try_int((result("td")[6]).text.replace(",", ""))
                                leechers = try_int((result("td")[7]).text.replace(",", ""))

                            except (AttributeError, TypeError):
                                continue

                            if not title or not download_url:
                                continue

                            # Filter unseeded torrent
                            if seeders < self.minseed or leechers < self.minleech:
                                if mode != "RSS":
                                    logger.debug(
                                        "Discarding torrent because it doesn't meet the minimum seeders or leechers: {0} (S:{1} L:{2})".format(
                                            title, seeders, leechers
                                        )
                                    )
                                continue

                            item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": ""}
                            if mode != "RSS":
                                logger.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))

                            items.append(item)

                except Exception:
                    logger.exception("Failed parsing provider. Traceback: {0}".format(traceback.format_exc()))

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)

            results += items

        return results



import os
import re

import bencodepy
from requests.utils import add_dict_to_cookiejar

from sickchill import logger, settings
from sickchill.oldbeard import helpers, tvcache
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class TorrentRssProvider(TorrentProvider):
    def __init__(self, name, url, cookies="", titleTAG="title", search_mode="eponly", search_fallback=False, enable_daily=False, enable_backlog=False):

        super().__init__(name)

        self.cache = TorrentRssCache(self, min_time=15)
        self.url = url.rstrip("/")

        self.supports_backlog = False

        self.search_mode = search_mode
        self.search_fallback = search_fallback
        self.enable_daily = enable_daily
        self.enable_backlog = enable_backlog
        self.enable_cookies = True
        self.cookies = cookies
        self.titleTAG = titleTAG

    def configStr(self):
        return "{0}|{1}|{2}|{3}|{4}|{5}|{6}|{7}|{8}".format(
            self.name or "",
            self.url or "",
            self.cookies or "",
            self.titleTAG or "",
            int(self.enabled),
            self.search_mode or "",
            int(self.search_fallback),
            int(self.enable_daily),
            int(self.enable_backlog),
        )

    @staticmethod
    def providers_list(data):
        providers_list = [x for x in (TorrentRssProvider._make_provider(x) for x in data.split("!!!")) if x]
        seen_values = set()
        providers_set = []

        for provider in providers_list:
            value = provider.name

            if value not in seen_values:
                providers_set.append(provider)
                seen_values.add(value)

        return [x for x in providers_set if x]

    def image_name(self):
        if os.path.isfile(os.path.join(settings.PROG_DIR, "gui", settings.GUI_NAME, "images", "providers", self.get_id() + ".png")):
            return self.get_id() + ".png"
        return "torrentrss.png"

    def _get_title_and_url(self, item):

        title = item.get(self.titleTAG, "").replace(" ", ".")

        attempt_list = [lambda: item.get("torrent_magneturi"), lambda: item.enclosures[0].href, lambda: item.get("link")]

        url = None
        for cur_attempt in attempt_list:
            try:
                url = cur_attempt()
            except Exception:
                continue

            if title and url:
                break

        return title, url

    @staticmethod
    def _make_provider(config):
        if not config:
            return None

        cookies = None
        enable_backlog = 0
        enable_daily = 0
        search_fallback = 0
        search_mode = "eponly"
        title_tag = "title"

        try:
            values = config.split("|")

            if len(values) == 9:
                name, url, cookies, title_tag, enabled, search_mode, search_fallback, enable_daily, enable_backlog = values
            elif len(values) == 8:
                name, url, cookies, enabled, search_mode, search_fallback, enable_daily, enable_backlog = values
            else:
                enabled = values[4]
                name = values[0]
                url = values[1]
        except ValueError:
            logger.exception("Skipping RSS Torrent provider string: {0}, incorrect format".format(config))
            return None

        new_provider = TorrentRssProvider(
            name,
            url,
            cookies=cookies,
            titleTAG=title_tag,
            search_mode=search_mode,
            search_fallback=search_fallback,
            enable_daily=enable_daily,
            enable_backlog=enable_backlog,
        )
        new_provider.enabled = enabled == "1"

        return new_provider

    def validateRSS(self):

        try:
            if self.cookies:
                success, status = self.add_cookies_from_ui()
                if not success:
                    return False, status

            # Access to a protected member of a client class
            data = self.cache._get_rss_data()["entries"]
            if not data:
                return False, "No items found in the RSS feed {0}".format(self.url)

            title, url = self._get_title_and_url(data[0])

            if not title:
                return False, "Unable to get title from first item"

            if not url:
                return False, "Unable to get torrent url from first item"

            if url.startswith("magnet:") and re.search(r"urn:btih:([\w]{32,40})", url):
                return True, "RSS feed Parsed correctly"
            else:
                torrent_file = self.get_url(url, returns="content")
                try:
                    bencodepy.decode(torrent_file)
                except (bencodepy.exceptions.BencodeDecodeError, Exception) as error:
                    self.dumpHTML(torrent_file)
                    return False, "Torrent link is not a valid torrent file: {0}".format(error)

            return True, "RSS feed Parsed correctly"

        except Exception as error:
            return False, "Error when trying to load RSS: {0}".format(str(error))

    @staticmethod
    def dumpHTML(data):
        dumpName = os.path.join(settings.CACHE_DIR, "custom_torrent.html")

        try:
            fileOut = open(dumpName, "wb")
            fileOut.write(data)
            fileOut.close()
            helpers.chmodAsParent(dumpName)
        except IOError as error:
            logger.exception("Unable to save the file: {0}".format(str(error)))
            return False

        logger.info("Saved custom_torrent html dump {0} ".format(dumpName))
        return True


class TorrentRssCache(tvcache.TVCache):
    def _get_rss_data(self):
        if self.provider.cookies:
            add_dict_to_cookiejar(self.provider.session.cookies, dict(x.rsplit("=", 1) for x in self.provider.cookies.split(";")))

        return self.get_rss_feed(self.provider.url)


Provider = TorrentRssProvider



from feedparser import parse

from .. import logger


def getFeed(url, params=None, request_hook=None):
    try:
        data = request_hook(url, params=params, returns="text", timeout=30)
        if not data:
            raise Exception

        feed = parse(data, response_headers={"content-type": "application/xml"})
        if feed:
            if "entries" in feed:
                return feed
            elif "error" in feed.feed:
                err_code = feed.feed["error"]["code"]
                err_desc = feed.feed["error"]["description"]
                logger.debug("RSS ERROR:[{0}] CODE:[{1}]".format(err_desc, err_code))
        else:
            logger.debug("RSS error loading data: " + url)

    except Exception as e:
        logger.debug("RSS error: " + str(e))

    return {"entries": []}



import json

import requests

from sickchill import logger, settings
from sickchill.oldbeard import common


class Notifier(object):

    SLACK_WEBHOOK_URL = "https://hooks.slack.com/services/"
    SLACK_ICON_URL = "https://github.com/SickChill/SickChill/raw/master/sickchill/gui/slick/images/sickchill-sc.png"

    def notify_snatch(self, ep_name):
        if settings.SLACK_NOTIFY_SNATCH:
            self._notify_slack(common.notifyStrings[common.NOTIFY_SNATCH] + ": " + ep_name)

    def notify_download(self, ep_name):
        if settings.SLACK_NOTIFY_DOWNLOAD:
            self._notify_slack(common.notifyStrings[common.NOTIFY_DOWNLOAD] + ": " + ep_name)

    def notify_subtitle_download(self, ep_name, lang):
        if settings.SLACK_NOTIFY_SUBTITLEDOWNLOAD:
            self._notify_slack(common.notifyStrings[common.NOTIFY_SUBTITLE_DOWNLOAD] + " " + ep_name + ": " + lang)

    def notify_git_update(self, new_version="??"):
        if settings.USE_SLACK:
            update_text = common.notifyStrings[common.NOTIFY_GIT_UPDATE_TEXT]
            title = common.notifyStrings[common.NOTIFY_GIT_UPDATE]
            self._notify_slack(title + " - " + update_text + new_version)

    def notify_login(self, ipaddress=""):
        if settings.USE_SLACK:
            update_text = common.notifyStrings[common.NOTIFY_LOGIN_TEXT]
            title = common.notifyStrings[common.NOTIFY_LOGIN]
            self._notify_slack(title + " - " + update_text.format(ipaddress))

    def test_notify(self):
        return self._notify_slack("This is a test notification from SickChill", force=True)

    def _send_slack(self, message=None):
        slack_webhook = self.SLACK_WEBHOOK_URL + settings.SLACK_WEBHOOK.replace(self.SLACK_WEBHOOK_URL, "")
        slack_icon_emoji = settings.SLACK_ICON_EMOJI

        logger.info("Sending slack message: " + message)
        logger.info("Sending slack message  to url: " + slack_webhook)

        headers = {"Content-Type": "application/json"}
        try:
            r = requests.post(
                slack_webhook,
                data=json.dumps(dict(text=message, username="SickChillBot", icon_emoji=slack_icon_emoji, icon_url=self.SLACK_ICON_URL)),
                headers=headers,
            )
            r.raise_for_status()
        except Exception as e:
            logger.exception("Error Sending Slack message: " + str(e))
            return False

        return True

    def _notify_slack(self, message="", force=False):
        if not settings.USE_SLACK and not force:
            return False

        return self._send_slack(message)



import os.path
import re
import sqlite3
import threading
import time
import traceback
import warnings
from sqlite3 import OperationalError
from typing import List

import sickchill.oldbeard.helpers
from sickchill import logger, settings

db_cons = {}
db_locks = {}


def db_full_path(filename="sickchill.db", suffix=None):
    """
    @param filename: The sqlite database filename to use. If not specified,
                     will be made to be sickchill.db
    @param suffix: The suffix to append to the filename. A "." will be added
                   automatically, i.e. suffix="v0" will make filename.db.v0
    @return: the correct location of the database file.
    """
    if suffix:
        filename = "{0}.{1}".format(filename, suffix)
    return os.path.join(settings.DATA_DIR, filename)


class DBConnection(object):
    MAX_ATTEMPTS = 5

    def __init__(self, filename="sickchill.db", suffix=None, row_type=None):

        self.filename = filename
        self.suffix = suffix
        self.row_type = row_type
        self.full_path = db_full_path(self.filename, self.suffix)

        if filename == "sickchill.db" and not os.path.isfile(self.full_path):
            sickbeard_db = db_full_path("sickbeard.db", suffix)
            if os.path.isfile(sickbeard_db):
                os.rename(sickbeard_db, self.full_path)

        try:
            if self.filename not in db_cons or not db_cons[self.filename]:
                db_locks[self.filename] = threading.Lock()

                self.connection = sqlite3.connect(self.full_path, 20, check_same_thread=False)
                db_cons[self.filename] = self.connection
            else:
                self.connection = db_cons[self.filename]

            # start off row factory configured as before out of
            # paranoia but wait to do so until other potential users
            # of the shared connection are done using
            # it... technically not required as row factory is reset
            # in all the public methods after the lock has been
            # acquired
            with db_locks[self.filename]:
                self._set_row_factory()

        except OperationalError:
            # noinspection PyUnresolvedReferences
            logger.warning(_("Please check your database owner/permissions: {db_filename}").format(db_filename=self.full_path))
        except Exception as e:
            self._error_log_helper(e, logger.ERROR, locals(), None, "DBConnection.__init__")
            raise

    def _error_log_helper(self, exception, severity, local_variables, attempts, called_method):
        if attempts in (0, self.MAX_ATTEMPTS):  # Only log the first try and the final failure
            prefix = ("Database", "Fatal")[severity == logger.ERROR]
            # noinspection PyUnresolvedReferences
            logger.log(
                severity,
                _("{exception_severity} error executing query with {method} in database {db_location}: ").format(
                    db_location=self.full_path, method=called_method, exception_severity=prefix
                )
                + str(exception),
            )

            # Lets print out all of the arguments so we can debug this better
            logger.info(traceback.format_exc())
            logger.info(_("If this happened in cache.db, you can safely stop SickChill, and delete the cache.db file without losing any data"))
            logger.info(_(f"Here are the arguments that were passed to this function (This is what the developers need to know): {local_variables}"))

    @staticmethod
    def _is_locked_or_denied(exception):
        # noinspection PyUnresolvedReferences
        return (
            _("unable to open database file") in exception.args[0]
            or _("database is locked") in exception.args[0]
            or "unable to open database file" in exception.args[0]
            or "database is locked" in exception.args[0]
        )

    def _set_row_factory(self):
        """
        once lock is acquired we can configure the connection for
        this particular instance of DBConnection
        """
        if self.row_type == "dict":
            self.connection.row_factory = DBConnection._dict_factory
        else:
            self.connection.row_factory = sqlite3.Row

    def _execute(self, query, args=None, fetchall=False, fetchone=False):
        """
        Executes DB query

        :param query: Query to execute
        :param args: Arguments in query
        :param fetchall: Boolean to indicate all results must be fetched
        :param fetchone: Boolean to indicate one result must be fetched (to walk results for instance)
        :return: query results
        """
        try:
            if not args:
                sql_results = self.connection.cursor().execute(query)
            else:
                sql_results = self.connection.cursor().execute(query, args)
            if fetchall:
                return sql_results.fetchall()
            elif fetchone:
                return sql_results.fetchone()
            else:
                return sql_results
        except Exception:
            raise

    def get_db_version(self):
        """
        Fetch major database version

        :return: Integer indicating current DB major version
        """
        if self.has_column("db_version", "db_minor_version"):
            warnings.warn("Deprecated: Use the version property", DeprecationWarning)
        return self.get_db_major_version()

    def get_db_major_version(self):
        """
        Fetch database version

        :return: Integer indicating current DB version
        """
        # noinspection PyBroadException
        try:
            result = int(self.select_one("SELECT db_version FROM db_version")[0])
            return result
        except Exception:
            return 0

    def get_db_minor_version(self):
        """
        Fetch database version

        :return: Integer indicating current DB major version
        """
        # noinspection PyBroadException
        try:
            result = int(self.select_one("SELECT db_minor_version FROM db_version")[0])
            return result
        except Exception:
            print("ERROR GETTING MINOR!")
            return 0

    @property
    def version(self):
        """The database version

        :return: A tuple containing the major and minor versions
        """
        # return tuple(self.select_one("SELECT * FROM db_version"))
        return self.get_db_major_version(), self.get_db_minor_version()

    def mass_action(self, query_list=None, log_transaction=False, fetchall=False):
        # type: (list, bool, bool) -> List[sqlite3.Row]
        """
        Execute multiple queries

        :param query_list: list of queries
        :param log_transaction: Boolean to wrap all in one transaction
        :param fetchall: Boolean, when using a select query force returning all results
        :return: list of results
        """

        # noinspection PyUnresolvedReferences
        assert hasattr(query_list, "__iter__"), _("You passed a non-iterable to mass_action: {0!r}").format(query_list)

        # remove None types
        query_list = [i for i in query_list if i]

        sql_results = []
        attempt = 0

        with db_locks[self.filename]:
            self._set_row_factory()
            while attempt <= self.MAX_ATTEMPTS:
                try:
                    log_level = (logger.DB, logger.DEBUG)[log_transaction]
                    for qu in query_list:
                        if len(qu) == 1:
                            # noinspection PyUnresolvedReferences
                            logger.log(log_level, _("{filename}: {query}").format(filename=self.filename, query=qu[0]))
                            sql_results.append(self._execute(qu[0], fetchall=fetchall))
                        elif len(qu) > 1:
                            # noinspection PyUnresolvedReferences
                            logger.log(log_level, _("{filename}: {query} with args {args}").format(filename=self.filename, query=qu[0], args=qu[1]))
                            sql_results.append(self._execute(qu[0], qu[1], fetchall=fetchall))
                    self.connection.commit()
                    # noinspection PyUnresolvedReferences
                    logger.log(log_level, _("Transaction with {count:d} of queries executed successfully").format(count=len(query_list)))

                    # finished
                    break
                except (sqlite3.OperationalError, sqlite3.DatabaseError) as e:
                    sql_results = []  # Reset results because of rollback
                    if self.connection:
                        self.connection.rollback()
                    severity = (logger.ERROR, logger.WARNING)[self._is_locked_or_denied(e) and attempt < self.MAX_ATTEMPTS]
                    self._error_log_helper(e, severity, locals(), attempt, "db.mass_action")
                    if severity == logger.ERROR:
                        raise
                    time.sleep(1)
                except Exception as e:
                    sql_results = []
                    if self.connection:
                        self.connection.rollback()
                    self._error_log_helper(e, logger.ERROR, locals(), attempt, "db.mass_action")
                    raise

                attempt += 1
            return sql_results

    def action(self, query, args=None, fetchall=False, fetchone=False):
        """
        Execute single query

        :param query: Query string
        :param args: Arguments to query string
        :param fetchall: Boolean to indicate all results must be fetched
        :param fetchone: Boolean to indicate one result must be fetched (to walk results for instance)
        :return: query results
        """
        if query is None:
            return

        # noinspection PyUnresolvedReferences
        assert not (fetchall and fetchone), _("Cannot fetch all and only one at the same time!")

        sql_results = []
        attempt = 0

        with db_locks[self.filename]:
            self._set_row_factory()
            while attempt < self.MAX_ATTEMPTS:
                try:
                    if settings.DBDEBUG:
                        if args is None:
                            logger.log(logger.DB, self.filename + ": " + query)
                        else:
                            logger.log(logger.DB, "{filename}: {query} with args {args}".format(filename=self.filename, query=query, args=args))

                    sql_results = self._execute(query, args, fetchall=fetchall, fetchone=fetchone)
                    self.connection.commit()

                    # get out of the connection attempt loop since we were successful
                    break
                except (sqlite3.OperationalError, sqlite3.DatabaseError) as e:
                    sql_results = []  # Reset results because of rollback
                    if self.connection:
                        self.connection.rollback()

                    severity = (logger.ERROR, logger.WARNING)[self._is_locked_or_denied(e) and attempt < self.MAX_ATTEMPTS]
                    self._error_log_helper(e, severity, locals(), attempt, "db.action")
                    if severity == logger.ERROR:
                        raise
                    time.sleep(1)
                except Exception as e:
                    sql_results = []
                    if self.connection:
                        self.connection.rollback()
                    self._error_log_helper(e, logger.ERROR, locals(), attempt, "db.action")
                    raise

                attempt += 1

            return sql_results

    def select(self, query, args=None):
        """
        Perform single select query on database

        :param query: query string
        :param args:  arguments to query string
        :return: query results
        """

        sql_results = self.action(query, args, fetchall=True)

        if sql_results is None:
            return []

        return sql_results

    def select_one(self, query, args=None):
        """
        Perform single select query on database, returning one result

        :param query: query string
        :param args: arguments to query string
        :return: query results
        """
        sql_results = self.action(query, args, fetchone=True)

        if sql_results is None:
            return []

        return sql_results

    def mass_upsert(self, table_name, query_list, log_transaction=False):
        # type: (str, List[tuple, list], bool) -> None
        """
        Execute multiple queries

        :param table_name: name of table to upsert
        :param query_list: list of queries
        :param log_transaction: Boolean to wrap all in one transaction
        :return: None
        """
        log_level = (logger.DB, logger.DEBUG)[log_transaction]
        for values, control in query_list:
            logger.log(log_level, _("{filename}: {query} [{control}]").format(filename=self.filename, query=values, control=control))
            self.upsert(table_name, values, control)

    def upsert(self, table_name, value_dict, key_dict):
        """
        Update values, or if no updates done, insert values
        TODO: Make this return true/false on success/error

        :param table_name: table to update/insert
        :param value_dict: values in table to update/insert
        :param key_dict:  columns in table to update/insert
        """

        changesBefore = self.connection.total_changes

        assert None not in list(key_dict.values()), _("Control dict to upsert cannot have values of None!")
        if key_dict:

            def make_string(my_dict, separator):
                return separator.join(["{} = ?".format(x) for x in my_dict])

            # language=TEXT
            query = "UPDATE [{table}] SET {pairs} WHERE {control}".format(
                table=table_name, pairs=make_string(value_dict, ", "), control=make_string(key_dict, " AND ")
            )

            self.action(query, list(value_dict.values()) + list(key_dict.values()))

        if self.connection.total_changes == changesBefore:
            keys = list(value_dict) + list(key_dict)
            count = len(keys)
            columns = ", ".join(keys)
            replacements = ", ".join(["?"] * count)
            values = list(value_dict.values()) + list(key_dict.values())

            # language=TEXT
            query = "INSERT INTO '{table}' ({columns}) VALUES ({replacements})".format(table=table_name, columns=columns, replacements=replacements)

            self.action(query, values)

    def table_info(self, table_name):
        """
        Return information on a database table

        :param table_name: name of table
        :return: array of name/type info
        """
        return {column["name"]: {"type": column["type"]} for column in self.select("PRAGMA table_info(`{0}`)".format(table_name))}

    @staticmethod
    def _dict_factory(cursor, row):
        return {col[0]: row[idx] for idx, col in enumerate(cursor.description)}

    def has_table(self, table_name):
        """
        Check if a table exists in database

        :param table_name: table name to check
        :return: True if table exists, False if it does not
        """
        return len(self.select("SELECT 1 FROM sqlite_master WHERE name = ?;", (table_name,))) > 0

    def has_column(self, table_name, column):
        """
        Check if a table has a column

        :param table_name: Table to check
        :param column: Column to check for
        :return: True if column exists, False if it does not
        """
        return column in self.table_info(table_name)

    def has_index(self, index_name):
        """
        Check if a table has an index

        :param index_name: Index to check
        :return: True if column exists, False if it does not
        """
        return len(self.select('SELECT 1 FROM sqlite_master WHERE type = "index" AND name = ?', [index_name])) > 0

    def add_column(self, table, column, col_type="NUMERIC", default=0):
        """
        Adds a column to a table, default column type is NUMERIC
        TODO: Make this return true/false on success/failure

        :param table: Table to add column too
        :param column: Column name to add
        :param col_type: Column type to add
        :param default: Default value for column
        """

        # language=TEXT
        self.action("ALTER TABLE [{0}] ADD {1} {2}".format(table, column, col_type))
        # language=TEXT
        self.action("UPDATE [{0}] SET {1} = ?".format(table, column), (default,))


def sanity_check_database(connection, sanity_check):
    sanity_check(connection).check()


class DBSanityCheck(object):
    def __init__(self, connection):
        self.connection = connection

    def check(self):
        pass


# ===============
# = Upgrade API =
# ===============


def upgrade_database(connection, schema):
    """
    Perform database upgrade and provide logging

    :param connection: Existing DB Connection to use
    :param schema: New schema to upgrade to
    """
    logger.debug("Checking database structure..." + connection.filename)
    _process_upgrade(connection, schema)


def pretty_name(class_name):
    return " ".join([x.group() for x in re.finditer("([A-Z])([a-z0-9]+)", class_name)])


def restore_database(version):
    """
    Restores a database to a previous version (backup file of version must still exist)

    :param version: Version to restore to
    :return: True if restore succeeds, False if it fails
    """
    logger.info("Restoring database before trying upgrade again")
    if not sickchill.oldbeard.helpers.restoreVersionedFile(db_full_path(suffix="v" + str(version)), version):
        logger.log_error_and_exit("Database restore failed, abort upgrading database")
        return False
    else:
        return True


def _process_upgrade(connection, upgrade_class):
    instance = upgrade_class(connection)
    # logger.debug("Checking " + pretty_name(upgrade_class.__name__) + " database upgrade")
    if not instance.test():
        logger.debug("Database upgrade required: " + pretty_name(upgrade_class.__name__))
        try:
            instance.execute()
        except Exception as e:
            logger.exception("Error in " + str(upgrade_class.__name__) + ": " + str(e))
            raise

        logger.debug(upgrade_class.__name__ + " upgrade completed")
    # else:
    #     logger.debug(upgrade_class.__name__ + " upgrade not required")

    for upgradeSubClass in upgrade_class.__subclasses__():
        _process_upgrade(connection, upgradeSubClass)


# Base migration class. All future DB changes should be subclassed from this class
class SchemaUpgrade(object):
    def __init__(self, connection):
        self.connection = connection

    def has_table(self, table_name):
        return self.connection.has_table(table_name=table_name)

    def has_column(self, table_name, column):
        return self.connection.has_column(table_name=table_name, column=column)

    def has_index(self, index_name):
        return self.connection.has_index(index_name=index_name)

    def add_column(self, table, column, col_type="NUMERIC", default=0):
        self.connection.add_column(table=table, column=column, col_type=col_type, default=default)

    def get_db_version(self):
        return self.connection.get_db_version()

    def increment_db_version(self):
        new_version = self.get_db_version() + 1
        self.connection.action("UPDATE db_version SET db_version = ?", [new_version])
        return new_version

    def inc_major_version(self):
        major_version, minor_version = self.connection.version
        major_version += 1
        minor_version = 0
        self.connection.action("UPDATE db_version SET db_version = ?, db_minor_version = ?", [major_version, minor_version])
        return self.connection.version

    def inc_minor_version(self):
        major_version, minor_version = self.connection.version
        minor_version += 1
        self.connection.action("UPDATE db_version SET db_version = ?, db_minor_version = ?", [major_version, minor_version])
        return self.connection.version



from putiopy import Client as PutioClient, ClientError

from sickchill.oldbeard import helpers
from sickchill.oldbeard.clients.generic import GenericClient


class Client(GenericClient):
    def __init__(self, host=None, username=None, password=None):

        super().__init__("put.io", host, username, password)
        self.url = "https://api.put.io/login"

    def _get_auth(self):
        client = PutioClient(self.password)
        try:
            client.Account.info()
        except ClientError as error:
            helpers.handle_requests_exception(error)
            self.auth = None
        else:
            self.auth = client

        return self.auth

    @property
    def _parent_id(self):
        parent_id = 0
        if self.username is not None and self.username != "":
            for f in self.auth.File.list():
                if f.name == self.username:
                    parent_id = f.id
                    break

        return parent_id

    def _add_torrent_uri(self, result):
        transfer = self.auth.Transfer.add_url(result.url, self._parent_id)

        return transfer.id is not None

    def _add_torrent_file(self, result):
        filename = result.name + ".torrent"
        transfer = self.auth.Transfer.add_torrent(filename, self._parent_id)

        return transfer.id is not None



import argparse
import gettext
import logging
import os
import site
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import List, Union
from urllib.request import urlopen

logger = logging.getLogger(__file__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())

sickchill_module = Path(__file__).parent.resolve()
pyproject_path = sickchill_module.parent / "pyproject.toml"
# locale_dir = sickchill_dir / "locale"
pid_file: Path = None


def sickchill_dir():
    return os.path.abspath(os.path.dirname(__file__))


def locale_dir():
    return os.path.abspath(os.path.join(os.path.dirname(__file__), "locale"))


BASE_PREFIX = getattr(sys, "base_prefix", getattr(sys, "real_prefix", sys.prefix))
IS_VIRTUALENV = BASE_PREFIX != sys.prefix


def setup_gettext(language: str = None) -> None:
    languages = [language] if language else None
    if not [key for key in ("LANGUAGE", "LC_ALL", "LC_MESSAGES", "LANG") if os.environ.get(key)]:
        os.environ["LC_MESSAGES"] = "en_US.UTF-8"
    gt = gettext.translation("messages", locale_dir(), languages=languages, fallback=True)
    gt.install(names=["ngettext"])


def maybe_daemonize():
    """
    Fork off as a daemon
    """
    if sys.platform == "win32":
        return

    args = argparse.ArgumentParser(add_help=False)
    args.add_argument("-d", "--daemon", action="store_true")
    args.add_argument("--pidfile")
    args, extra = args.parse_known_args(sys.argv)

    if not args.daemon:
        return

    global pid_file
    if args.pidfile:
        pid_file = Path(args.pidfile).resolve()

    if pid_file:
        if pid_file.is_file():
            # If the pid file already exists, SickChill may still be running, so exit
            raise SystemExit(f"PID file: {pid_file} already exists. Exiting.")
        pid_dir = pid_file.parent
        if not os.access(pid_dir, os.F_OK):
            raise SystemExit(f"PID dir: {pid_dir} doesn't exist. Exiting.")
        if not os.access(pid_dir, os.W_OK):
            raise SystemExit(f"PID dir: {pid_dir} must be writable (write permissions). Exiting.")

    # An object is accessed for a non-existent member.
    # Access to a protected member of a client class
    # Make a non-session-leader child process
    try:
        pid = os.fork()  # @UndefinedVariable - only available in UNIX
        if pid != 0:
            os._exit(0)
    except OSError as error:
        raise SystemExit(f"fork #1 failed: {error.errno}: {error.strerror}\n")

    os.setsid()  # @UndefinedVariable - only available in UNIX

    # https://github.com/SickChill/SickChill/issues/2969
    # http://www.microhowto.info/howto/cause_a_process_to_become_a_daemon_in_c.html#idp23920
    # https://www.safaribooksonline.com/library/view/python-cookbook/0596001673/ch06s08.html
    # Previous code simply set the umask to whatever it was because it was ANDing instead of OR-ing
    # Daemons traditionally run with umask 0 anyways and this should not have repercussions
    os.umask(0)

    # Make the child a session-leader by detaching from the terminal
    try:
        pid = os.fork()
        if pid != 0:
            os._exit(0)
    except OSError as error:
        raise SystemExit(f"fork #2 failed: {error.errno}: {error.strerror}\n")

    # Write pid
    if pid_file:
        pid = os.getpid()

        logger.info(f"Writing PID: {pid} to {pid_file}\n")
        try:
            with open(pid_file, "w") as pid_fd:
                pid_fd.write(f"{pid}\n")
        except EnvironmentError as error:
            raise SystemExit(f"Unable to write PID file: {pid_file} Error {error.errno}: {error.strerror}")

    # Redirect all output
    sys.stdout.flush()
    sys.stderr.flush()

    devnull = getattr(os, "devnull", "/dev/null")
    stdin = open(devnull)
    stdout = open(devnull, "a+")
    stderr = open(devnull, "a+")

    os.dup2(stdin.fileno(), getattr(sys.stdin, "device", sys.stdin).fileno())
    os.dup2(stdout.fileno(), getattr(sys.stdout, "device", sys.stdout).fileno())
    os.dup2(stderr.fileno(), getattr(sys.stderr, "device", sys.stderr).fileno())


def remove_pid_file():
    """
    Remove pid file

    :return:
    """
    try:
        if pid_file and pid_file.exists():
            pid_file.unlink()
    except EnvironmentError:
        pass


def check_installed(name: str = __package__) -> bool:
    try:
        from importlib.metadata import Distribution, PackageNotFoundError  # noqa
    except ImportError:
        try:
            from importlib_metadata import Distribution, PackageNotFoundError  # noqa
        except ImportError:
            # Should not get here EVER, but just in case lets just try checking pip freeze instead
            result, output = subprocess.getstatusoutput([f"{sys.executable} -m pip freeze"])
            if result != 0:  # Not Ok
                return False
            is_installed = name in [requirement.split("==")[0] for requirement in output.splitlines()]
            logger.debug(f"{name} found: {is_installed}")
            return is_installed

    try:
        Distribution.from_name(name)
        logger.debug(f"{name} found: True")
    except PackageNotFoundError:
        logger.debug(f"{name} found: False")
        return False
    return True


def subprocess_call(cmd_list):
    try:
        process = subprocess.Popen(cmd_list, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
        stdout, stderr = process.communicate()
        process.wait()
        if stdout or stderr:
            logger.info(f"Command result: {stdout or stderr}")
    except Exception as error:
        logger.info(f"Unable to run command: {error}")
        return 126
    return process.returncode


def get_os_id():
    os_release = Path("/etc/os-release").resolve()
    if os_release.is_file():
        from configparser import ConfigParser

        parser = ConfigParser()
        parser.read_string("[DEFAULT]\n" + os_release.read_text())
        try:
            return parser["DEFAULT"]["ID"]
        except (KeyError, IndexError):
            pass


def pip_install(packages: Union[List[str], str]) -> bool:
    if not isinstance(packages, list):
        packages = packages.splitlines()

    cmd = [
        sys.executable,
        "-m",
        "pip",
        "install",
        "--no-input",
        "--disable-pip-version-check",
        "--no-python-version-warning",
        "--no-color",
        "--trusted-host=pypi.org",
        "--trusted-host=files.pythonhosted.org",
        "-qU",
    ]

    os_id = get_os_id()
    if os_id in ("alpine", "ubuntu"):
        cmd.append(f"--find-links=https://wheel-index.linuxserver.io/{os_id}/")

    cmd += packages

    logger.debug(f"pip args: {' '.join(cmd)}")

    result = subprocess_call(cmd)
    if result != 0:  # Not Ok
        logger.info("Trying user site-packages")
        result = subprocess_call(cmd + ["--user"])
        if result != 0:  # Not Ok
            return False
    return True


def check_env_writable():
    """
    Checks if we can install packages to the current environment
    """
    locations = []
    try:
        locations = [site.getsitepackages()[0]]
    except (AttributeError, IndexError):
        pass

    try:
        if site.ENABLE_USER_SITE or site.check_enableusersite():
            if site.USER_SITE:
                locations.append(site.USER_SITE)
            else:
                locations.append(site.getusersitepackages())
    except AttributeError:
        pass

    if not locations:
        try:
            from distutils.sysconfig import get_python_lib

            locations = [get_python_lib()]
        except (ImportError, ModuleNotFoundError):
            pass

    if not locations:
        try:
            import pip

            locations = [str(Path(pip.__path__[0]).parent.resolve())]
        except (ImportError, AttributeError, IndexError):
            pass

    return any([os.access(location, os.W_OK) for location in locations])


def download_to_temp_file(url: str) -> tempfile.NamedTemporaryFile:
    filename = url.rsplit("/", 1)[-1]
    prefix, suffix = filename.rsplit(".", 1)
    tfd = tempfile.NamedTemporaryFile(delete=False, suffix="." + suffix, prefix=prefix, mode="wb")
    tfd.write(urlopen(url).read())
    tfd.close()
    return tfd


def check_and_install_pip() -> None:
    """
    Downloads and runs get-pip.py, installing pip in the current environment
    """

    if not check_installed("pip"):
        logger.info("Installing pip")
        tfd = download_to_temp_file("https://bootstrap.pypa.io/get-pip.py")
        result = subprocess_call([f"{sys.executable}", f"{tfd.name}"])
        os.unlink(tfd.name)
        if result == 0:
            logger.info("Pip installed")
        else:
            logger.info("There was an error installing pip!")


def make_virtualenv_and_rerun(location: Path) -> None:
    """
    This is a hail mary, when we cannot install to the already existing virtualenv because
    someone created it as root or another user than the user running this process

    Creates a .venv dir in project root (which is gitignored already)
    if found SC will restart automatically using tthe environment in .venv
    """

    location = location.resolve()
    current_interpreter = Path(sys.executable).resolve()

    result = 0  # Ok
    if str(location) == str(sys.prefix):
        if IS_VIRTUALENV:
            logger.info(f"Unable to install to the existing virtual environment located at {sys.prefix}")
            logger.info("Please check the permissions, and that it does not include global site packages")
        result = 126  # Command invoked cannot execute
    else:
        if not location.is_dir():
            logger.info(f"Because of the above errors, we will try creating a new virtualenvironment in {location}")
            if not check_installed("virtualenv"):
                logger.info("virtualenv module not found, getting a portable one to use temporarily")
                tfd = download_to_temp_file("https://bootstrap.pypa.io/virtualenv.pyz")
                result = subprocess_call([f"{sys.executable}", f"{tfd.name}", "-p", f"{sys.executable}", f"{location}"])
                os.unlink(tfd.name)
                if result != 0:  # Not Ok
                    logger.info("Due to the above error, we cannot continue! Exiting")
                else:
                    logger.info(f"Created new virtualenvironment in {location}")
            else:
                result = subprocess_call([f"{sys.executable}", "-m", "virtualenv", "-p", f"{sys.executable}", f"{location}"])
                if result != 0:  # Not Ok
                    logger.info("Due to the above error, we cannot continue! Exiting")
                else:
                    logger.info(f"Created new virtualenvironment in {location}")

        if location.is_dir() and result == 0:  # Ok
            locations_to_check = []
            # append the bin/python.ext to the new venv path

            check = location
            for part in current_interpreter.parts[-2:]:
                if sys.platform == "win32" and part == "tools":
                    part = "Scripts"
                check /= part
            locations_to_check.append(check)

            locations_to_check.append(location / "bin" / current_interpreter.parts[-1])
            locations_to_check.append(location / "Scripts" / current_interpreter.parts[-1])

            locations_to_check.extend(x for x in location.rglob("*python3.?") if x.is_file())
            locations_to_check.extend(x for x in location.rglob("*python3") if x.is_file())
            locations_to_check.extend(x for x in location.rglob("*python") if x.is_file())

            for place in locations_to_check:
                if place.is_file() and place.stat().st_mode & os.X_OK:
                    # add original arguments to this re-call
                    new_argv = [str(place)] + sys.argv

                    logger.info(f"Restarting SickChill with {new_argv}")
                    return os.execvp(new_argv[0], new_argv)

            logger.info(f"Something weird happend when creating the virtualenv, Could not find the bin dir or executable in {location}. Exiting")

    os._exit(result)


def poetry_install() -> None:
    logger.info("Checking poetry")
    if not check_installed():
        if pyproject_path.exists():
            # Check if we can write to this virtualenv
            if not check_env_writable():
                logger.info(f"Current environment is not writable!")
                if not os.access(pyproject_path.parent, os.W_OK):
                    logger.info(f"Source dir is not writable by this user either, we cannot continue: f{pyproject_path.parent}")
                    os._exit(126)

                make_virtualenv_and_rerun(pyproject_path.with_name(".venv"))

            if not IS_VIRTUALENV:
                logger.info(f"We always run from virtualenv when running from source")
                make_virtualenv_and_rerun(pyproject_path.with_name(".venv"))

            check_and_install_pip()

            # Cool, we can write to site-packages
            pip_install(["setuptools", "poetry", "--pre"])
            if check_installed("poetry"):
                result, output = subprocess.getstatusoutput(
                    f"cd {pyproject_path.parent} && {sys.executable} -m poetry export -f requirements.txt --without-hashes"
                )
                if result == 0:  # Ok
                    pip_install(output)
                else:  # Not Ok
                    logger.info(output)
                    make_virtualenv_and_rerun(pyproject_path.with_name(".venv"))
            else:  # Couldn't install poetry, make new venv
                make_virtualenv_and_rerun(pyproject_path.with_name(".venv"))



from urllib.parse import urljoin

from sickchill import logger
from sickchill.helper.exceptions import AuthException
from sickchill.oldbeard import tvcache
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        super().__init__("Shazbat.tv")

        self.supports_backlog = False

        self.passkey = None
        self.options = None

        self.cache = ShazbatCache(self, min_time=20)

        self.url = "http://www.shazbat.tv"
        self.urls = {
            "login": urljoin(self.url, "login"),
            "rss_recent": urljoin(self.url, "rss/recent"),
            # 'rss_queue': urljoin(self.url, 'rss/download_queue'),
            # 'rss_followed': urljoin(self.url, 'rss/followed')
        }

    def _check_auth(self):
        if not self.passkey:
            raise AuthException("Your authentication credentials are missing, check your config.")

        return True

    def _check_auth_from_data(self, data):
        if not self.passkey:
            self._check_auth()
        elif data.get("bozo") == 1 and not (data["entries"] and data["feed"]):
            logger.warning("Invalid username or password. Check your settings")

        return True


class ShazbatCache(tvcache.TVCache):
    def _get_rss_data(self):
        params = {"passkey": self.provider.passkey, "fname": "true", "limit": 100, "duration": "2 hours"}

        return self.get_rss_feed(self.provider.urls["rss_recent"], params=params)

    def _check_auth(self, data):
        return self.provider._check_auth_from_data(data)



import os
from urllib.parse import urlencode
from urllib.request import Request, urlopen

import requests

from sickchill import logger, settings, tv


class Notifier(object):
    def notify_snatch(self, ep_name):
        pass

    def notify_download(self, ep_name):
        pass

    def notify_subtitle_download(self, ep_name, lang):
        pass

    def notify_git_update(self, new_version):
        pass

    def notify_login(self, ipaddress=""):
        pass

    @staticmethod
    def update_library(ep_obj):

        # Values from config

        if not settings.USE_PYTIVO:
            return False

        host = settings.PYTIVO_HOST
        shareName = settings.PYTIVO_SHARE_NAME
        tsn = settings.PYTIVO_TIVO_NAME

        # There are two more values required, the container and file.
        #
        # container: The share name, show name and season
        #
        # file: The file name
        #
        # Some slicing and dicing of variables is required to get at these values.
        #
        # There might be better ways to arrive at the values, but this is the best I have been able to
        # come up with.
        #

        # Calculated values
        # noinspection PyUnresolvedReferences
        if isinstance(ep_obj, tv.TVEpisode):
            showPath = ep_obj.show.location
            showName = ep_obj.show.name
            rootShowAndSeason = os.path.dirname(ep_obj.location)
            absPath = ep_obj.location
        else:
            # This is a TVShow
            showPath = ep_obj.location
            showName = ep_obj.name
            rootShowAndSeason = os.path.dirname(ep_obj.location)
            absPath = ep_obj.location

        # Some show names have colons in them which are illegal in a path location, so strip them out.
        # (Are there other characters?)
        showName = showName.replace(":", "")

        root = showPath.replace(showName, "")
        showAndSeason = rootShowAndSeason.replace(root, "")

        container = shareName + "/" + showAndSeason
        filename = "/" + absPath.replace(root, "")

        # Finally create the url and make request
        requestUrl = "http://" + host + "/TiVoConnect?" + urlencode({"Command": "Push", "Container": container, "File": filename, "tsn": tsn})

        logger.debug("pyTivo notification: Requesting " + requestUrl)

        request = Request(requestUrl)

        try:
            urlopen(request)
        except requests.exceptions.RequestException as e:
            if hasattr(e, "reason"):
                logger.exception("pyTivo notification: Error, failed to reach a server - " + e.reason)
                return False
            elif hasattr(e, "code"):
                logger.exception("pyTivo notification: Error, the server couldn't fulfill the request - " + e.code)
            return False
        except Exception as e:
            logger.exception("PYTIVO: Unknown exception: " + str(e))
            return False
        else:
            logger.info("pyTivo notification: Successfully requested transfer of file")
            return True
